# pylint: disable=W0212, W0201
# W0212: Access to a protected member of a client class
# W0201: Attribute defined outside __init__

from ftw.pdfgenerator.builder import Builder
from ftw.pdfgenerator.exceptions import BuildTerminated, PDFBuildFailed
from ftw.pdfgenerator.interfaces import IBuilder, IBuilderFactory
from ftw.pdfgenerator.testing import PREDEFINED_BUILD_DIRECTORY_LAYER
from ftw.testing import MockTestCase
from mocker import MATCH, ANY
from StringIO import StringIO
from zipfile import ZipFile
from zope.component import getUtility
from zope.interface.verify import verifyClass
import os


class TestBuilder(MockTestCase):

    layer = PREDEFINED_BUILD_DIRECTORY_LAYER

    def setUp(self):
        super(TestBuilder, self).setUp()
        self.builddir = self.layer.builddir

    def test_utility_is_factory(self):
        factory = getUtility(IBuilderFactory)
        self.assertFalse(IBuilder.implementedBy(factory))
        self.assertEqual(factory().__class__, Builder)

    def test_factory_returns_builder(self):
        factory = getUtility(IBuilderFactory)
        builder = factory()
        self.assertTrue(IBuilder.providedBy(builder))
        verifyClass(IBuilder, Builder)

    def test_builder_implements_interface(self):
        self.assertTrue(IBuilder.implementedBy(Builder))
        verifyClass(IBuilder, Builder)

    def test_build_directory_is_mocked_in_test_setup(self):
        factory = getUtility(IBuilderFactory)
        builder = factory()
        self.assertEqual(builder.build_directory, self.builddir)

    def test_add_file(self):
        builder = getUtility(IBuilderFactory)()
        builder.add_file('foo.txt', 'Foo\nBar')

        self.assertTrue(os.path.exists(self.builddir))
        filepath = os.path.join(self.builddir, 'foo.txt')
        self.assertTrue(os.path.exists(filepath),
                        'File not found: {0}'.format(filepath))
        self.assertEqual(open(filepath).read(), 'Foo\nBar')

    def test_add_file_by_file_object(self):
        builder = getUtility(IBuilderFactory)()
        builder.add_file('foo.txt', StringIO('Foo\nBar'))

        self.assertTrue(os.path.exists(self.builddir))
        filepath = os.path.join(self.builddir, 'foo.txt')
        self.assertTrue(os.path.exists(filepath),
                        'File not found: {0}'.format(filepath))
        self.assertEqual(open(filepath).read(), 'Foo\nBar')

    def test_terminated_exception_raised(self):
        builder = self.mocker.patch(getUtility(IBuilderFactory)())
        fake_pdf_path = os.path.join(self.builddir, 'export.pdf')
        fake_pdf = open(fake_pdf_path, 'w')
        fake_pdf.write('the pdf')
        fake_pdf.close()

        self.expect(builder._build_pdf(ANY)).result(fake_pdf_path)
        self.replay()

        builder.build('LaTeX')

        with self.assertRaises(BuildTerminated) as cm:
            builder.add_file('foo.txt', 'Foo\nBar')
        self.assertEqual(str(cm.exception),
                         'The build is already terminated.')

        with self.assertRaises(BuildTerminated) as cm:
            builder.build('LaTeX')
        self.assertEqual(str(cm.exception),
                         'The build is already terminated.')

        with self.assertRaises(BuildTerminated) as cm:
            builder.build_zip('LaTeX')
        self.assertEqual(str(cm.exception),
                         'The build is already terminated.')

    def test_build_removes_directory(self):
        builder = self.mocker.patch(getUtility(IBuilderFactory)())

        def execute_mock(cmd):
            self.assertEqual(
                open(os.path.join(self.builddir, 'export.tex')).read(),
                'LaTeX Code')

            pdf = open(os.path.join(self.builddir, 'export.pdf'), 'w')
            pdf.write('Rendered PDF')
            pdf.close()
            return 0, '', ''

        self.expect(
            builder._execute(
                MATCH(lambda cmd: cmd.startswith('pdflatex ')))
            ).call(execute_mock)

        self.replay()

        self.assertTrue(builder.config.remove_build_directory)
        self.assertEqual(builder.build('LaTeX Code'), 'Rendered PDF')
        self.assertFalse(os.path.exists(builder.build_directory))

    def test_build_does_not_remove_directory_if_disabled(self):
        builder = self.mocker.patch(getUtility(IBuilderFactory)())
        self.expect(builder.config.remove_build_directory).result(False)

        def execute_mock(cmd):
            self.assertEqual(
                open(os.path.join(self.builddir, 'export.tex')).read(),
                'LaTeX Code')

            pdf = open(os.path.join(self.builddir, 'export.pdf'), 'w')
            pdf.write('Rendered PDF')
            pdf.close()
            return 0, '', ''

        self.expect(
            builder._execute(
                MATCH(lambda cmd: cmd.startswith('pdflatex ')))
            ).call(execute_mock)

        self.replay()

        self.assertEqual(builder.build('LaTeX Code'), 'Rendered PDF')
        self.assertTrue(os.path.exists(builder.build_directory))

    def test_build_zip_returns_valid_zip_file_as_stream(self):
        builder = self.mocker.patch(getUtility(IBuilderFactory)())

        def execute_mock(cmd):
            self.assertEqual(
                open(os.path.join(self.builddir, 'export.tex')).read(),
                'LaTeX Code')

            pdf = open(os.path.join(self.builddir, 'export.pdf'), 'w')
            pdf.write('Rendered PDF')
            pdf.close()
            return 0, '', ''

        self.expect(
            builder._execute(
                MATCH(lambda cmd: cmd.startswith('pdflatex ')))
            ).call(execute_mock)

        self.replay()
        self.assertTrue(builder.config.remove_build_directory)

        builder.add_file('mystyle.sty', 'LaTeX sty file content')

        data = builder.build_zip('LaTeX Code')
        self.assertTrue(hasattr(data, 'read'))

        zipobj = ZipFile(data)
        filenames_in_zip = zipobj.namelist()

        self.assertIn('export.tex', filenames_in_zip)
        self.assertIn('export.pdf', filenames_in_zip)
        self.assertIn('mystyle.sty', filenames_in_zip)

        self.assertEqual('LaTeX Code', zipobj.read('export.tex'))
        self.assertEqual('LaTeX sty file content',
                         zipobj.read('mystyle.sty'))

        self.assertFalse(os.path.exists(builder.build_directory))

    def test_build_removes_directory_even_if_build_failed(self):
        builder = self.mocker.patch(getUtility(IBuilderFactory)())

        self.expect(builder._execute(ANY)).result((
                1, '', 'could not build pdf for some reason...'))

        self.replay()

        with self.assertRaises(PDFBuildFailed) as cm:
            builder.build('LaTeX')
        self.assertEqual(str(cm.exception),
                         'PDF missing.')

        self.assertFalse(os.path.exists(builder.build_directory))

    def test_build_zip_removes_directory_even_if_build_failed(self):
        builder = self.mocker.patch(getUtility(IBuilderFactory)())

        self.expect(builder._execute(ANY)).result((
                1, '', 'could not build pdf for some reason...'))

        self.replay()

        builder.build_zip('LaTeX')

        self.assertFalse(os.path.exists(builder.build_directory))

    def test_execute_runs_command_in_build_directory(self):
        builder = getUtility(IBuilderFactory)()

        file_ = open(os.path.join(builder.build_directory, 'foo.txt'), 'w+')
        file_.write('bar')
        file_.close()

        self.assertEqual(builder._execute('ls .'), (0, 'foo.txt\n', ''))
        self.assertEqual(builder._execute('cat foo.txt'), (0, 'bar', ''))
        self.assertEqual(builder._execute('cat baz'),
                         (1, '', 'cat: baz: No such file or directory\n'))

    def test_build_pdf_reruns_pdflatex_if_needed(self):
        builder = self.mocker.patch(getUtility(IBuilderFactory)())
        aux_path = os.path.join(self.builddir, 'export.aux')
        pdf_path = os.path.join(self.builddir, 'export.pdf')

        self._reruns__executed_call_counter = 0

        def exec_mock(cmd):
            self._reruns__executed_call_counter += 1
            if self._reruns__executed_call_counter == 1:
                aux = open(aux_path, 'w+')
                aux.write('first run')
                aux.close()
                pdf = open(pdf_path, 'w+')
                pdf.write('the pdf')
                pdf.close()
                return (0, 'some log\nRerun to get it better\ndone', '')

            elif self._reruns__executed_call_counter == 2:
                aux = open(aux_path, 'wa')
                aux.write('second run')
                aux.close()
                return (0, 'some log', '')

            elif self._reruns__executed_call_counter == 3:
                # no change in aux
                return (0, 'some log', '')

        self.expect(builder._execute(ANY)).call(exec_mock).count(3)

        self.replay()

        builder._build_pdf(u'The latex')

    def test_maximum_reruns(self):
        builder = getUtility(IBuilderFactory)()
        rerun_limit = builder._rerun_limit
        builder = self.mocker.patch(builder)

        aux_path = os.path.join(self.builddir, 'export.aux')

        def exec_mock(cmd):
            aux = open(aux_path, 'a+')
            aux.write('another run')
            aux.close()
            return (0, 'some log', '')

        self.expect(builder._execute(ANY)).call(exec_mock).count(rerun_limit)

        self.replay()

        with self.assertRaises(PDFBuildFailed) as cm:
            builder._build_pdf('The latex')
        self.assertEqual(str(cm.exception),
                         'Maximum pdf build limit reached.')

    def test_build_pdf_executes_makeindex(self):
        builder = self.mocker.patch(getUtility(IBuilderFactory)())
        aux_path = os.path.join(self.builddir, 'export.aux')
        pdf_path = os.path.join(self.builddir, 'export.pdf')

        def pdflatex_call_mock(cmd):
            with open(aux_path, 'w+') as aux:
                aux.write('first run')
            with open(pdf_path, 'w+') as pdf:
                pdf.write('the pdf')
            return (0, 'the log', '')

        # Runs 2 times by default, does not rerun because _makeindex
        # returns False (non rerun required)
        self.expect(builder._execute(ANY)).call(pdflatex_call_mock).count(2)
        self.expect(builder._makeindex()).result(False)

        self.replay()

        builder._build_pdf(u'LaTeX')

    def test_build_pdf_reruns_when_makeindex_requires_rerun(self):
        builder = self.mocker.patch(getUtility(IBuilderFactory)())
        aux_path = os.path.join(self.builddir, 'export.aux')
        pdf_path = os.path.join(self.builddir, 'export.pdf')

        def pdflatex_call_mock(cmd):
            with open(aux_path, 'w+') as aux:
                aux.write('first run')
            with open(pdf_path, 'w+') as pdf:
                pdf.write('the pdf')
            return (0, 'the log', '')

        # Runs 3 times: 2 by default + once because _makeindex requires
        # an additional run by returning True
        self.expect(builder._execute(ANY)).call(pdflatex_call_mock).count(3)
        self.expect(builder._makeindex()).result(True)

        self.replay()

        builder._build_pdf(u'LaTeX')

    def test_makeindex_does_nothing_and_returns_False_without_idx_file(self):
        builder = getUtility(IBuilderFactory)()
        self.assertEquals(False, builder._makeindex())

    def test_makeindex_calls_executable_and_returns_True_with_idx_file(self):
        builder = self.mocker.patch(getUtility(IBuilderFactory)())
        idx_path = os.path.join(self.builddir, 'export.idx')
        with open(idx_path, 'w+') as idx:
            idx.write('\indexentry{Test}{2}')

        (self.expect(builder._execute('makeindex -g -s umlaut.ist export'))
         .result((0, 'stdout', 'stderr')))

        self.replay()
        self.assertEquals(True, builder._makeindex())

    def test_cleanup(self):
        builder = getUtility(IBuilderFactory)()
        self.assertTrue(os.path.exists(builder.build_directory))
        builder.cleanup()
        self.assertFalse(os.path.exists(builder.build_directory))



from StringIO import StringIO
from ftw.pdfgenerator.exceptions import BuildTerminated, PDFBuildFailed
from ftw.pdfgenerator.interfaces import IBuilder, IConfig
from zipfile import ZipFile
from zope.component import getUtility
from zope.interface import implements
import os
import shlex
import shutil
import subprocess


RESOURCES_DIR = os.path.abspath(os.path.join(__file__, '..', 'resources'))


class Builder(object):
    implements(IBuilder)

    def __init__(self):
        self.config = getUtility(IConfig)
        self.build_directory = self.config.get_build_directory()
        self._terminated = False
        self._aux_data = None
        self._rerun_limit = 10

    def add_file(self, filename, data):
        if self._terminated:
            raise BuildTerminated('The build is already terminated.')

        path = os.path.join(self.build_directory, filename)
        with open(path, 'wb') as fio:
            if hasattr(data, 'read'):
                shutil.copyfileobj(data, fio)
            else:
                fio.write(data)

    def build(self, latex):
        if self._terminated:
            raise BuildTerminated('The build is already terminated.')

        try:
            pdf_path = self._build_pdf(latex)

        except PDFBuildFailed:
            self._cleanup_build()
            raise

        else:
            pdf_file = open(pdf_path, 'rb')
            data = pdf_file.read()
            pdf_file.close()
            self._cleanup_build()
            return data

    def build_zip(self, latex):
        if self._terminated:
            raise BuildTerminated('The build is already terminated.')

        try:
            self._build_pdf(latex)

        except PDFBuildFailed:
            pass
        data = StringIO()
        zip_file = ZipFile(data, 'w')

        for filename in os.listdir(self.build_directory):
            zip_file.write(os.path.join(self.build_directory, filename),
                      filename)

        zip_file.close()
        data.seek(0)

        self._cleanup_build()
        return data

    def cleanup(self):
        if not self._terminated:
            self._cleanup_build()

    def _build_pdf(self, latex):
        """Build the pdf in the build_directory and return the path to
        the pdf.
        """
        latex_path = os.path.join(self.build_directory, 'export.tex')
        pdf_path = os.path.join(self.build_directory, 'export.pdf')

        assert not os.path.exists(latex_path), 'export.tex already exists'
        assert not os.path.exists(pdf_path), 'export.pdf already exists'

        latex_file = open(latex_path, 'w')
        if isinstance(latex, unicode):
            latex = latex.encode('utf-8')
        latex_file.write(latex)
        latex_file.close()

        self._run_pdflatex(latex_path)
        if self._makeindex():
            self._run_pdflatex(latex_path)

        if not os.path.exists(pdf_path):
            raise PDFBuildFailed('PDF missing.')

        return pdf_path

    def _run_pdflatex(self, latex_path):
        self._aux_data = None
        cmd = 'pdflatex --interaction=nonstopmode %s' % latex_path
        stdout = ''
        while self._rerun_required(stdout):
            _exitcode, stdout, _stderr = self._execute(cmd)

    def _makeindex(self):
        idx_path = os.path.join(self.build_directory, 'export.idx')
        if not os.path.exists(idx_path):
            return False

        # When the makeindex arguments are too long, we might get a
        # buffer oferflow.
        # For avoiding this, we copy the "umlaut.ist" to the export directory.
        umlaut_ist_path = os.path.join(RESOURCES_DIR, 'umlaut.ist')
        shutil.copyfile(umlaut_ist_path, os.path.join(self.build_directory, 'umlaut.ist'))
        self._execute('makeindex -g -s umlaut.ist export')
        return True

    def _rerun_required(self, stdout):
        if self._rerun_limit == 0:
            raise PDFBuildFailed('Maximum pdf build limit reached.')

        self._rerun_limit -= 1

        previous_aux_data = self._aux_data

        self._aux_data = []
        for filename in os.listdir(self.build_directory):
            if filename.endswith('.aux'):
                path = os.path.join(self.build_directory, filename)
                file_ = open(path)
                self._aux_data.append(file_.read())
                file_.close()

        if 'Rerun to get' in stdout:
            return True

        elif previous_aux_data is None:
            return True

        elif previous_aux_data != self._aux_data:
            return True

        else:
            return False

    def _cleanup_build(self):
        self._terminated = True

        if self.config.remove_build_directory:
            shutil.rmtree(self.build_directory)

    def _execute(self, cmd):
        proc = subprocess.Popen(shlex.split(cmd),
                                stderr=subprocess.PIPE,
                                stdout=subprocess.PIPE,
                                cwd=self.build_directory)
        output, errors = proc.communicate()
        return proc.poll(), output, errors


def builder_factory():
    return Builder()



# pylint: disable=W0212, W0201
# W0212: Access to a protected member of a client class
# W0201: Attribute defined outside __init__

from ftw.pdfgenerator.html2latex.converter import HTML2LatexConverter
from ftw.testing import MockTestCase


class TestURLConverter(MockTestCase):

    def setUp(self):
        super(TestURLConverter, self).setUp()

        self.context = self.mocker.mock(count=False)
        self.expect(self.context.absolute_url()).result(
            'http://nohost/plone')

        self.layout = self.mocker.mock()

        self.converter = HTML2LatexConverter(
            context=self.context,
            request=object(),
            layout=self.layout)

        self.convert = self.converter.convert

    def test_simple_url(self):
        self.replay()
        html = 'x http://domain.ch/foo/bar y'
        latex = 'x http://domain.ch""/foo""/bar y'
        self.assertEqual(self.convert(html), latex)

    def test_https_url(self):
        self.replay()
        html = 'x https://domain.ch/foo/bar y'
        latex = 'x https://domain.ch""/foo""/bar y'
        self.assertEqual(self.convert(html), latex)

    def test_advanced_url(self):
        self.replay()
        html = 'x http://usr@pwd:sub.domain.com/path/to/doc.html' + \
               '?foo=1&bar=2#anchor y'
        latex = 'x http://usr@pwd:sub.domain.com""/path""/to""/doc.html' + \
                '?foo=1&bar=2\\#anchor y'
        self.assertEqual(self.convert(html), latex)



from ftw.pdfgenerator.html2latex.converter import HTML2LatexConverter
from ftw.pdfgenerator.html2latex.subconverter import SubConverter
from ftw.pdfgenerator.interfaces import ISubConverter
from ftw.testing import MockTestCase
from zope.interface.verify import verifyClass
import re


class TestBaseSubConverter(MockTestCase):

    def test_implements_interface(self):
        self.assertTrue(ISubConverter.implementedBy(SubConverter))
        verifyClass(ISubConverter, SubConverter)

    def test_init(self):
        converter = object()
        match = object()
        html = object()

        obj = SubConverter(converter=converter, match=match, html=html)
        self.assertEqual(obj.converter, converter)
        self.assertEqual(obj.match, match)
        self.assertEqual(obj.fullhtml, html)

    def test_call_not_implemented(self):
        converter = match = html = object()
        obj = SubConverter(converter, match, html)

        with self.assertRaises(NotImplementedError):
            obj()

    def test_get_html_returns_matched_html(self):
        html = 'one three five'
        match = re.search('t[\w]*', html)

        obj = SubConverter(object(), match, html)
        self.assertEqual(obj.get_html(), 'three')

    def test_replace_and_lock_passed_to_converter(self):
        html = 'one three five'
        match = re.search('t[\w]*', html)
        converter = self.mocker.mock()

        self.expect(converter.replace_and_lock(4, 9, 'latex code'))

        self.replay()

        obj = SubConverter(converter, match, html)
        obj.replace_and_lock('latex code')

    def test_replace_passed_to_converter(self):
        html = 'one three five'
        match = re.search('t[\w]*', html)
        converter = self.mocker.mock()

        self.expect(converter.replace(4, 9, 'latex code'))

        self.replay()

        obj = SubConverter(converter, match, html)
        obj.replace('latex code')

    def test_get_context(self):
        context = self.mocker.mock()
        self.expect(context.found())

        class MySubConverter(SubConverter):
            pattern = 'y'

            def __call__(self):
                self.get_context().found()

        self.replay()

        converter = HTML2LatexConverter(
            context=context,
            request=object(),
            layout=object())

        converter.convert('xyz', custom_subconverters=[MySubConverter])

    def test_get_layout(self):
        layout = self.mocker.mock()
        self.expect(layout.found())

        class MySubConverter(SubConverter):
            pattern = 'y'

            def __call__(self):
                self.get_layout().found()

        self.replay()

        converter = HTML2LatexConverter(
            context=object(),
            request=object(),
            layout=layout)

        converter.convert('xyz', custom_subconverters=[MySubConverter])



from ftw.pdfgenerator import babel
from ftw.testing import MockTestCase


class TestBabel(MockTestCase):

    def test_lookup_babel_option(self):
        self.assertEqual(babel.lookup_babel_option('en'), 'english')
        self.assertEqual(babel.lookup_babel_option('en-gb'), 'british')
        self.assertEqual(babel.lookup_babel_option('en-us'), 'USenglish')
        self.assertEqual(babel.lookup_babel_option('en-gm'), 'english')

    def test_get_preferred_babel_option_for_context(self):
        foo = self.mocker.mock()
        self.expect(foo.getLanguage).result(lambda: 'de-at')

        language_tool = self.mocker.mock()
        self.mock_tool(language_tool, 'portal_languages')
        self.expect(language_tool.getPreferredLanguage()).result('de')

        bar = self.mocker.mock()
        self.expect(getattr(bar, 'getLanguage', None)).result(None)

        self.replay()

        self.assertEqual(babel.get_preferred_babel_option_for_context(foo),
                         'naustrian')
        self.assertEqual(babel.get_preferred_babel_option_for_context(bar),
                         'ngerman')



from ftw.pdfgenerator.layout.baselayout import BaseLayout
from ftw.pdfgenerator.templating import MakoTemplating


class MakoLayoutBase(BaseLayout, MakoTemplating):
    """A base layout with enabled templating engine.

    When subclassing, define `template_directories` with a list of relative
    paths to your template directories.
    Define the default template by setting `template_name` to the filename.
    The default template should be in one of your template directories.
    """

    template_directories = []
    template_name = None

    def __init__(self, context, request, builder):
        BaseLayout.__init__(self, context, request, builder)
        MakoTemplating.__init__(self)

    def render_latex(self, content_latex):
        """Renders the template `template_name`, which is in one of the
        defined `template_directories`.
        """

        if self.template_name is None:
            raise ValueError('%s: `template_name` is not defined.' % (
                    self.__class__.__name__))

        self.before_render_hook()

        args = self.get_render_arguments()
        args['content'] = content_latex
        args['packages'] = self.get_packages_latex()

        return self.render_layout_template(args)

    def render_layout_template(self, args):
        return self.render_template(self.template_name, **args)

    def get_render_arguments(self):
        """Returns a dict of arguments passed to the default template.
        """
        return {}

    def before_render_hook(self):
        """This method is called before rendering and may register
        required packages (`use_package`).
        """
        pass

    def add_raw_template_file(self, name):
        """Adds the contents of a file to the builder without parsing it.
        """
        self.builder.add_file(name, data=self.get_raw_template(name))



# pylint: disable=W0212, W0201
# W0212: Access to a protected member of a client class
# W0201: Attribute defined outside __init__

from ftw.pdfgenerator.html2latex.converter import HTML2LatexConverter
from ftw.testing import MockTestCase


LATEX_HREF = r'\href{%(url)s}{%(label)s\footnote{\href{%(url)s}' + \
    r'{\url{%(url_label)s}}}}'


class TestHyperlinkConverter(MockTestCase):

    def setUp(self):
        super(TestHyperlinkConverter, self).setUp()

        self.context = self.mocker.mock(count=False)
        self.expect(self.context.absolute_url()).result(
            'http://nohost/plone')

        self.layout = self.mocker.mock()
        self.expect(self.layout.use_package('url', 'hyphens')).count(1, None)
        self.expect(self.layout.use_package('hyperref')).count(1, None)

        self.converter = HTML2LatexConverter(
            context=self.context,
            request=object(),
            layout=self.layout)

        self.convert = self.converter.convert

        self.reference_catalog = self.mocker.mock(count=False)
        self.mock_tool(self.reference_catalog, 'reference_catalog')

    def test_converts_urls(self):
        self.replay()
        html = '<a href="http://www.google.com/">guugel</a>'
        latex = LATEX_HREF % {'label': 'guugel',
                              'url': 'http://www.google.com/',
                              'url_label': 'http://www.google.com/'}
        self.assertEqual(self.convert(html), latex)

    def test_relative_urls(self):
        self.replay()
        html = '<a href="./foo/bar">baz</a>'
        latex = LATEX_HREF % {
            'label': 'baz',
            'url': 'http://nohost/plone/foo/bar',
            'url_label': 'http://nohost/plone/foo/bar'}
        self.assertEqual(self.convert(html), latex)

        html = '<a href="foo/bar">baz</a>'
        latex = LATEX_HREF % {
            'label': 'baz',
            'url': 'http://nohost/plone/foo/bar',
            'url_label': 'http://nohost/plone/foo/bar'}
        self.assertEqual(self.convert(html), latex)

    def test_links_within_brackets(self):
        self.replay()
        html = 'foo (<a href="http://google.com">bar</a>) baz'
        latex = r'foo (%s) baz' % LATEX_HREF % {
            'label': 'bar',
            'url': 'http://google.com',
            'url_label': 'http://google.com'}
        self.assertEqual(self.convert(html), latex)

    def test_hyphens_in_links(self):
        self.replay()
        html = '<a href="http://my-page.com">foo</a>'
        latex = LATEX_HREF % {'label': 'foo',
                              'url': 'http://my-page.com',
                              'url_label': 'http://my-page.com'}
        self.assertEqual(self.convert(html), latex)

    def test_mailto_links(self):
        self.replay()
        html = '<a href="mailto:info@google.com">foo</a>'
        latex = LATEX_HREF % {'label': 'foo',
                              'url': 'mailto:info@google.com',
                              'url_label': 'info@google.com'}
        self.assertEqual(self.convert(html), latex)

    def test_spaces_in_links(self):
        self.replay()
        html = '<a href="http://foo/foo%20bar%20baz">foo</a>'
        latex = LATEX_HREF % {
            'label': 'foo',
            'url': 'http://foo/foo\%20bar\%20baz',
            'url_label': 'http://foo/foo\%20bar\%20baz'}
        self.assertEqual(self.convert(html), latex)

        html = '<a href="mailto:info@google.com?subject=foo%20bar%20baz">' +\
            'foo</a>'
        latex = LATEX_HREF % {
            'label': 'foo',
            'url': 'mailto:info@google.com?subject=foo\%20bar\%20baz',
            'url_label': 'info@google.com?subject=foo\%20bar\%20baz'}
        self.assertEqual(self.convert(html), latex)

        html = '<a href="mailto:info@google.com?subject=foo bar baz">foo</a>'
        latex = LATEX_HREF % {
            'label': 'foo',
            'url': 'mailto:info@google.com?subject=foo\%20bar\%20baz',
            'url_label': 'info@google.com?subject=foo\%20bar\%20baz'}
        self.assertEqual(self.convert(html), latex)

    def test_ampersand_in_links(self):
        self.replay()
        html = '<a href="http://host.com/?foo=1&amp;bar=2">baz</a>'
        latex = LATEX_HREF % {'label': 'baz',
                              'url': 'http://host.com/?foo=1\\&bar=2',
                              'url_label': 'http://host.com/?foo=1\\&bar=2'}
        self.assertEqual(self.convert(html), latex)

        html = '<a href="http://host.com/?foo=1&bar=2">baz</a>'
        latex = LATEX_HREF % {'label': 'baz',
                              'url': 'http://host.com/?foo=1\\&bar=2',
                              'url_label': 'http://host.com/?foo=1\\&bar=2'}
        self.assertEqual(self.convert(html), latex)

    def test_underscores_in_links(self):
        self.replay()
        html = '<a href="http://host.com/foo_bar">baz</a>'
        latex = LATEX_HREF % {'label': 'baz',
                              'url': 'http://host.com/foo_bar',
                              'url_label': 'http://host.com/foo\\_bar'}
        self.assertEqual(self.convert(html), latex)

    def test_hash_key_in_links(self):
        self.replay()
        html = '<a href="http://host.com/foo#bar">baz</a>'
        latex = LATEX_HREF % {'label': 'baz',
                              'url': 'http://host.com/foo\\#bar',
                              'url_label': 'http://host.com/foo\\#bar'}
        self.assertEqual(self.convert(html), latex)

    def test_additional_parameters(self):
        self.replay()
        html = 'foo <a color="red" href="http://nohost/bar" class="link">' + \
            'bar</a> baz'
        latex = r'foo %s baz' % LATEX_HREF % {
            'label': 'bar',
            'url': 'http://nohost/bar',
            'url_label': 'http://nohost/bar'}

        self.assertEqual(self.convert(html), latex)

    def test_multiple_links(self):
        self.replay()
        html = '<a href="http://foo">foo</a> <a href="http://bar">bar</a> ' +\
            '<a href="http://baz">baz</a>'

        link1 = LATEX_HREF % {'label': 'foo',
                              'url': 'http://foo',
                              'url_label': 'http://foo'}

        link2 = LATEX_HREF % {'label': 'bar',
                              'url': 'http://bar',
                              'url_label': 'http://bar'}

        link3 = LATEX_HREF % {'label': 'baz',
                              'url': 'http://baz',
                              'url_label': 'http://baz'}

        latex = ' '.join((link1, link2, link3))

        self.assertEqual(self.convert(html), latex)

    def test_other_protocols(self):
        self.replay()
        protocols = ['http', 'https', 'ftp', 'file']

        for protocol in protocols:
            html = '<a href="%s://foo/bar">baz</a>' % protocol
            latex = LATEX_HREF % {'label': 'baz',
                                  'url': '%s://foo/bar' % protocol,
                                  'url_label': '%s://foo/bar' % protocol}
            self.assertEqual(self.convert(html), latex)

    def test_label_is_converted(self):
        self.replay()
        html = '<a href="http://foo">foo <b>bar</b> baz</a>'
        latex = LATEX_HREF % {'label': r'foo \textbf{bar} baz',
                              'url': 'http://foo',
                              'url_label': 'http://foo'}
        self.assertEqual(self.convert(html), latex)

    def test_no_escaped_dashes_in_label(self):
        # In a \href{} environment, a-b should not become a"=b because
        # this will be printend literal.
        self.replay()
        html = '<a href="http://foo">x-y</a>'
        latex = LATEX_HREF % {'label': r'x-y',
                              'url': 'http://foo',
                              'url_label': 'http://foo'}
        self.assertEqual(self.convert(html), latex)

    def test_removes_nonbreaking_spaces(self):
        self.replay()
        # Non break spaces are evil. In HTML they are usually not used the
        # way they should be used in LaTeX, so we replace them with spaces.
        html = u'<a href="http://host">Hello\xa0World</a>'.encode('utf8')
        latex = LATEX_HREF % {'label': 'Hello World',
                              'url': 'http://host',
                              'url_label': 'http://host'}
        self.assertEqual(self.convert(html), latex)

    def test_resolveuid_links_are_resolved(self):
        obj = self.context.stub()
        self.expect(obj.absolute_url()).result('http://nohost/theobj')

        self.expect(self.reference_catalog.lookupObject('THEUID')).result(obj)
        self.replay()

        html = '<a href="./resolveuid/THEUID">The Obj</a>'
        latex = LATEX_HREF % {'label': 'The Obj',
                              'url': 'http://nohost/theobj',
                              'url_label': 'http://nohost/theobj'}
        self.assertEqual(self.convert(html), latex)

        html = '<a href="./resolveUid/THEUID">The Obj</a>'
        latex = LATEX_HREF % {'label': 'The Obj',
                              'url': 'http://nohost/theobj',
                              'url_label': 'http://nohost/theobj'}
        self.assertEqual(self.convert(html), latex)

    def test_links_in_listing_items(self):
        self.replay()

        # There should not be a non-escaped ampersand (&) within listing
        # items - even when in a href.
        html = '<ol><li><a href="%s">foo bar</a></li></ol>' % (
            'http://host/view?foo=1&bar=2')

        latex_link = LATEX_HREF % {
            'label': 'foo bar',
            'url': 'http://host/view?foo=1\\&bar=2',
            'url_label': 'http://host/view?foo=1\\&bar=2'}

        latex = '\n'.join((
                r'\begin{enumerate}',
                r'\item %s' % latex_link,
                r'\end{enumerate}',
                ))

        self.assertEqual(self.convert(html).strip(), latex)



from ftw.pdfgenerator.html2latex.subconverters import htmlentities
from ftw.pdfgenerator.tests.base import SubconverterTestBase



class TestHtmlentitiesConverter(SubconverterTestBase):

    def test_converter_is_default(self):
        self.assertIn(
            htmlentities.HtmlentitiesConverter,
            self.converter.get_default_subconverters())

    def test_converter_converts_umlauts(self):
        self.assertEqual(self.convert('&auml;&euml;&iuml;&ouml;&uuml;'),
                         '\xc3\xa4\xc3\xab\xc3\xaf\xc3\xb6\xc3\xbc')
        self.assertEqual(self.convert('5 &lt; 6'),
                         '5 < 6')
        self.assertEqual(self.convert('Hello&nbsp;World'),
                         'Hello World')

    def test_converter_converts_xmlentities(self):
        # "&" should also be escaped by a later pattern.
        self.assertEqual(self.convert('m&#38;m'),
                         r'm\&m')



from ftw.pdfgenerator.html2latex.subconverters import textformatting
from ftw.pdfgenerator.tests.base import SubconverterTestBase



class TestTextformattingConverter(SubconverterTestBase):

    def test_converter_is_default(self):
        self.assertIn(
            textformatting.Textformatting,
            self.converter.get_default_subconverters())

    def test_find_bracket_range(self):
        text = 'foo {bar} baz'
        span = (4, 9)
        self.assertEqual(
            textformatting.find_bracket_range(text),
            span)
        self.assertEqual(text[span[0]:span[1]], '{bar}')

        self.assertEqual(
            textformatting.find_bracket_range(
                'one {two {three {four }}} five'),
            (4, 25))

        self.assertEqual(
            textformatting.find_bracket_range('} foo {bar} baz'),
            (6, 11))

        self.assertEqual(
            textformatting.find_bracket_range('1 } 2 { 3'),
            None)

    def test_newline_in_bold_command(self):
        # "\textbf{X\n\n}" is not allowed, use "\textbf{X\\}"
        self.assertEqual(
            self.convert(r'<p><b>foo<br /> <br /></b></p>'),
            r'\textbf{foo\\\\}')

    def test_newline_in_bold_command2(self):
        # "\textbf{X\n\nY}" is not allowed, use "\textbf{X\\Y}"
        self.assertEqual(
            self.convert(r'<p><b>foo<br /> <br />bar</b></p>'),
            r'\textbf{foo\\\\bar}')

    def test_newline_in_bold_command3(self):
        # "\textbf{\n\nY}" is not allowed, use "\textbf{\\Y}"
        self.assertEqual(
            self.convert(r'<b><br /> <br />bar</b>' * 2),
            r'\textbf{\\\\bar}' * 2)

    def test_newline_in_emph_command(self):
        # "\emph{X\n\n}" is not allowed, use "\emph{X\\}"
        self.assertEqual(
            self.convert(r'foo <u>bar<br /> <br /></u> baz'),
            r'foo \emph{bar\\\\} baz')

    def test_newline_in_emph_command2(self):
        # "\emph{X\n\nY}" is not allowed, use "\emph{X\\Y}"
        self.assertEqual(
            self.convert(r'<u>foo<br /> <br />bar</u>'),
            r'\emph{foo\\\\bar}')

    def test_newline_in_emph_command3(self):
        # "\emph{\n\nY}" is not allowed, use "\emph{\\Y}"
        self.assertEqual(
            self.convert(r'<u><br /> <br />bar</u>' * 2),
            r'\emph{\\\\bar}' * 2)

    def test_newline_in_italic_command(self):
        # "\textit{X\n\n}" is not allowed, use "\textit{X\\}"
        self.assertEqual(
            self.convert(r'<p><i>foo<br /> <br /></i></p>'),
            r'\textit{foo\\\\}')

    def test_newline_in_italic_command2(self):
        # "\textit{X\n\nY}" is not allowed, use "\textit{X\\Y}"
        self.assertEqual(
            self.convert(r'<p><i>foo<br /> <br />bar</i></p>'),
            r'\textit{foo\\\\bar}')

    def test_newline_in_italic_command3(self):
        # "\textit{\n\nY}" is not allowed, use "\textit{\\Y}"
        self.assertEqual(
            self.convert(r'<i><br /> <br />bar</i>' * 2),
            r'\textit{\\\\bar}' * 2)

    def test_newline_in_combined_italic_and_bold_command(self):
        # "\textit{*\n\n*}" is not allowed, use "\textit{*\\*}"
        # "\textbf{*\n\n*}" is not allowed, use "\textbf{*\\*}"
        self.assertEqual(
            self.convert(r'<i>foo'
                         r'<b>bar</b><br /> <br />'
                         r'<u>baz</u><br /> <br />'
                         r'</i>'),
            r'\textit{foo\textbf{bar}\\\\\emph{baz}\\\\}')

        self.assertEqual(
            self.convert(r'<b>foo<i>bar</i><br /> <br /></b>'),
            r'\textbf{foo\textit{bar}\\\\}')

    def test_newline_only_in_textish_environments_changed(self):
        self.assertEqual(
            self.convert(r'<p><i>foo</i><br /><br ><p><p>bar</p>' * 2),
            ('\\textit{foo}\n\nbar\n\n' * 2).strip())

    def test_multiple_textish_environments(self):
        self.assertEqual(
            self.convert(
                r'<p>one <b>two</b> three</p><p>four <b>five</b></p>'),
            'one \\textbf{two} three\n\nfour \\textbf{five}')



# pylint: disable=W0212, W0201
# W0212: Access to a protected member of a client class
# W0201: Attribute defined outside __init__

from ftw.pdfgenerator import interfaces
from ftw.pdfgenerator.html2latex import converter
from ftw.pdfgenerator.html2latex import subconverter
from ftw.pdfgenerator.html2latex import wrapper
from ftw.pdfgenerator.layout.baselayout import BaseLayout
from ftw.pdfgenerator.testing import PDFGENERATOR_ZCML_LAYER
from ftw.testing import MockTestCase
from unittest2 import TestCase
from zope.component import getMultiAdapter
from zope.interface.verify import verifyClass


MODE_REPLACE = interfaces.HTML2LATEX_MODE_REPLACE
MODE_REGEXP = interfaces.HTML2LATEX_MODE_REGEXP
MODE_REGEXP_FUNCTION = interfaces.HTML2LATEX_MODE_REGEXP_FUNCTION
DEFAULT_PLACEHOLDER = interfaces.HTML2LATEX_CUSTOM_PATTERN_PLACEHOLDER


class TestCustomPatternPlaceholderWrapper(TestCase):

    def test_placeholder_wrapper(self):
        placeholder = interfaces.HTML2LATEX_CUSTOM_PATTERN_PLACEHOLDER_BOTTOM
        mode = interfaces.HTML2LATEX_MODE_REGEXP_FUNCTION

        wrap = wrapper.CustomPatternAtPlaceholderWrapper(
            mode=mode,
            placeholder=placeholder)

        self.assertEqual(wrap.mode, mode)
        self.assertEqual(wrap.placeholder, placeholder)


class TestBasePatternAware(TestCase):

    def setUp(self):
        super(TestBasePatternAware, self).setUp()

        self.pattern1 = (MODE_REPLACE, '{', '\\{')
        self.pattern2 = (MODE_REPLACE, '}', '\\}')
        self.pattern3 = (MODE_REPLACE, '&nbsp;', ' ')
        self.pattern4 = (MODE_REPLACE, '\r\n', ' ')
        self.pattern5 = (MODE_REPLACE, '\r', ' ')
        self.pattern6 = (MODE_REPLACE, '\n', ' ')

    def test_register_patterns_at_default_placeholder(self):
        obj = converter.BasePatternAware(
            [self.pattern1,
             self.pattern2,
             DEFAULT_PLACEHOLDER,
             self.pattern3])

        obj.register_patterns([self.pattern4, self.pattern5])

        self.assertEqual(
            obj.patterns,
            [self.pattern1,
             self.pattern2,
             self.pattern4,
             self.pattern5,
             DEFAULT_PLACEHOLDER,
             self.pattern3,])

    def test_register_patterns_with_placeholder_wrapper(self):
        top = interfaces.HTML2LATEX_CUSTOM_PATTERN_PLACEHOLDER_TOP
        obj = converter.BasePatternAware(
            [top,
             self.pattern1,
             DEFAULT_PLACEHOLDER,
             self.pattern2])

        special_pattern = (
            wrapper.CustomPatternAtPlaceholderWrapper(
                mode=MODE_REGEXP,
                placeholder=top),
            r'\w{5, 7}', ' ')
        special_pattern_unwrapped = [MODE_REGEXP, r'\w{5, 7}', ' ']

        obj.register_patterns([special_pattern])

        self.assertEqual(
            obj.patterns,
            [special_pattern_unwrapped,
             top,
             self.pattern1,
             DEFAULT_PLACEHOLDER,
             self.pattern2])

    def test_register_patterns_overwrite_existing(self):
        obj = converter.BasePatternAware(
            [self.pattern1,
             DEFAULT_PLACEHOLDER])

        custom = list(self.pattern1)
        custom[2] = 'foo'
        custom = tuple(custom)

        obj.register_patterns([custom])

        self.assertEqual(
            obj.patterns,
            [custom,
             DEFAULT_PLACEHOLDER])

    def test_reigster_custom_pattern_with_noreplace(self):
        obj = converter.BasePatternAware([
                self.pattern1,
                DEFAULT_PLACEHOLDER])

        custom = list(self.pattern1)
        custom[2] = 'foo'
        custom = tuple(custom)

        obj._insert_custom_pattern(custom, replace=False)

        self.assertEqual(
            obj.patterns,
            [self.pattern1,
             custom,
             DEFAULT_PLACEHOLDER])

    def test_register_wrapper_replacing(self):
        top = interfaces.HTML2LATEX_CUSTOM_PATTERN_PLACEHOLDER_TOP
        obj = converter.BasePatternAware(
            [top,
             self.pattern1,
             DEFAULT_PLACEHOLDER])

        custom = [
            wrapper.CustomPatternAtPlaceholderWrapper(
                mode=self.pattern1[0],
                placeholder=top),
            self.pattern1[1],
            'foo']

        obj.register_patterns([custom])

        self.assertEqual(
            obj.patterns,
            [[self.pattern1[0], self.pattern1[1], 'foo'],
             top,
             self.pattern1,
             DEFAULT_PLACEHOLDER])

    def test_register_subconverters(self):
        obj = converter.BasePatternAware([DEFAULT_PLACEHOLDER])

        class Foo(subconverter.SubConverter):
            pattern = 'foo'
            placeholder = DEFAULT_PLACEHOLDER

            def __call__(self):
                return ''

        obj.register_subconverters([Foo])

        self.assertEqual(
            obj.patterns,
            [(MODE_REGEXP_FUNCTION, 'foo', Foo),
             DEFAULT_PLACEHOLDER])

    def test_get_subconverter_by_pattern(self):
        obj = converter.BasePatternAware([DEFAULT_PLACEHOLDER])

        class Foo(subconverter.SubConverter):
            pattern = 'foo'
            placeholder = DEFAULT_PLACEHOLDER

            def __call__(self):
                return ''

        class Bar(Foo):
            pattern = 'bar'

        class Baz(Foo):
            pattern = 'baz'

        obj.register_subconverters([Foo, Bar, Baz])
        self.assertEqual(obj.get_subconverter_by_pattern('foo'), Foo)
        self.assertEqual(obj.get_subconverter_by_pattern('bar'), Bar)
        self.assertEqual(obj.get_subconverter_by_pattern('baz'), Baz)
        self.assertEqual(obj.get_subconverter_by_pattern('unkown'), None)


class TestHTML2LatexConverter(MockTestCase):

    layer = PDFGENERATOR_ZCML_LAYER

    def test_implements_interface(self):
        self.assertTrue(interfaces.IHTML2LaTeXConverter.implementedBy(
                converter.HTML2LatexConverter))
        verifyClass(interfaces.IHTML2LaTeXConverter,
                    converter.HTML2LatexConverter)

    def test_default_converter_adapter_registration(self):
        context = self.create_dummy()
        request = self.create_dummy()
        builder = self.create_dummy()
        layout = BaseLayout(context, request, builder)

        obj = getMultiAdapter((context, request, layout),
                              interfaces.IHTML2LaTeXConverter)

        self.assertEqual(obj.__class__, converter.HTML2LatexConverter)

    def test_converter_uses_copy_of_default_subconverters(self):
        class NoSubconvertersConverter(converter.HTML2LatexConverter):
            def get_default_subconverters(self):
                return []

        obj = NoSubconvertersConverter(
            object(), object(), object())

        self.assertEqual(obj.patterns, converter.DEFAULT_PATTERNS)
        self.assertNotEqual(id(obj.patterns), id(converter.DEFAULT_PATTERNS))

    def test_convert_with_custom_pattern(self):
        obj = converter.HTML2LatexConverter(
            object(), object(), object())

        self.assertEqual(
            obj.convert('foo bar baz', custom_patterns=[
                    (MODE_REPLACE, 'bar', 'replaced')]),
            'foo replaced baz')

    def test_convert_with_custom_subconverter(self):
        class MyConverter(subconverter.SubConverter):
            pattern = 'bar'

            def __call__(self):
                self.replace(r'\textbf{replaced}')

        obj = converter.HTML2LatexConverter(
            object(), object(), object())

        self.assertEqual(
            obj.convert('foo bar baz bar', custom_subconverters=[
                    MyConverter]),
            r'foo \textbf{replaced} baz \textbf{replaced}')

    def test_convert_plain(self):
        obj = converter.HTML2LatexConverter(
            object(), object(), object())

        self.assertEqual(obj.convert('foo <bar> baz'), 'foo  baz')
        self.assertEqual(obj.convert_plain('foo <bar> baz'), 'foo <bar> baz')

    def test_quoted_umlaut(self):
        obj = converter.HTML2LatexConverter(
            object(), object(), object())

        self.assertEqual(
            '"Uber h"ofliche B"urger aus Rh\xc3\xb4ne-Alpes',
            obj.quoted_umlauts(
                '\xc3\x9cber h\xc3\xb6fliche B\xc3\xbcrger'
                ' aus Rh\xc3\xb4ne-Alpes'))


class TestHTML2LaTeXConvertRunner(MockTestCase):

    def test_implements_interface(self):
        self.assertTrue(interfaces.IHTML2LaTeXConvertRunner.implementedBy(
                converter.HTML2LatexConvertRunner))
        verifyClass(interfaces.IHTML2LaTeXConvertRunner,
                    converter.HTML2LatexConvertRunner)


    def test_raises_when_initializied_with_non_converter(self):
        with self.assertRaises(ValueError) as cm:
            converter.HTML2LatexConvertRunner(object(), [], '')

        self.assertEqual(str(cm.exception),
                         'converter should by a IHTML2LaTeXConverter')

    def test_raises_when_convert_started_twice(self):
        obj = converter.HTML2LatexConverter(
            object(), object(), object())
        runner = converter.HTML2LatexConvertRunner(obj, [], '')

        runner.runner_convert()
        with self.assertRaises(RuntimeError) as cm:
            runner.runner_convert()

        self.assertEqual(str(cm.exception),
                         'runner_convert() should not be called twice!')

    def test_quoted_umlaut(self):
        obj = converter.HTML2LatexConverter(
            object(), object(), object())
        runner = converter.HTML2LatexConvertRunner(obj, [], '')

        self.assertEqual(
            '"Uber h"ofliche B"urger aus Rh\xc3\xb4ne-Alpes',
            runner.quoted_umlauts(
                '\xc3\x9cber h\xc3\xb6fliche B\xc3\xbcrger'
                ' aus Rh\xc3\xb4ne-Alpes'))



from ftw.pdfgenerator.config import DefaultConfig
from ftw.pdfgenerator.interfaces import IConfig
from ftw.testing.layer import ComponentRegistryLayer
from mocker import Mocker, expect
from plone.mocktestcase.dummy import Dummy
from plone.testing import Layer
from zope.component import getGlobalSiteManager
from zope.component import provideUtility
from zope.component.hooks import setSite
from zope.i18n.interfaces import IUserPreferredLanguages
from zope.interface import alsoProvides
import os
import shutil
import tempfile


class PDFGeneratorZCMLLayer(ComponentRegistryLayer):
    """A layer which only sets up the zcml, but does not start a zope
    instance.
    """

    def setUp(self):
        super(PDFGeneratorZCMLLayer, self).setUp()
        import ftw.pdfgenerator
        self.load_zcml_file('test.zcml', ftw.pdfgenerator.tests)
        self.load_zcml_file('configure.zcml', ftw.pdfgenerator)


PDFGENERATOR_ZCML_LAYER = PDFGeneratorZCMLLayer()


class ZCMLWithSiteLayer(Layer):
    """A layer with loaded zcml and a global site.
    """

    defaultBases = (PDFGENERATOR_ZCML_LAYER,)

    def testSetUp(self):
        setSite(self._create_site_with_request())

    def testTearDown(self):
        setSite(None)

    def _create_site_with_request(self):
        request = Dummy(getPreferredLanguages=lambda: [])
        alsoProvides(request, IUserPreferredLanguages)
        site = Dummy(REQUEST=request,
                     getSiteManager=getGlobalSiteManager)

        return site


ZCML_WITH_SITE_LAYER = ZCMLWithSiteLayer()


class PredefinedBuildDirectoryLayer(Layer):

    defaultBases = (PDFGENERATOR_ZCML_LAYER,)

    def setUp(self):
        self.testcase_mocker = Mocker()
        self.builddir = os.path.join(tempfile.mkdtemp('test-builder'),
                                     'build')

        self.config = self.testcase_mocker.proxy(
            DefaultConfig(),
            spec=None,
            count=False)
        expect(self.config.get_build_directory()).result(self.builddir)

        self.testcase_mocker.replay()

    def testSetUp(self):
        provideUtility(provides=IConfig, component=self.config)
        os.mkdir(self.builddir)

    def tearDown(self):
        self.testcase_mocker.verify()
        self.testcase_mocker.restore()

    def testTearDown(self):
        if os.path.exists(self.builddir):
            shutil.rmtree(self.builddir)


PREDEFINED_BUILD_DIRECTORY_LAYER = PredefinedBuildDirectoryLayer()



from BeautifulSoup import BeautifulSoup
from ftw.pdfgenerator.html2latex import subconverter
from ftw.pdfgenerator.utils import html2xmlentities
from xml.dom import minidom
from xml.parsers.expat import ExpatError


# LaTeX allows a maximum list nesting of 4. Deeper nesting will be flattened
# to the nesting limit so that the produced LaTeX code is still valid.
LIST_NESTING_LIMIT = 4


class ListConverter(subconverter.SubConverter):
    """
    The ListConverter converts <ul> and <ol> lists
    into latex' itemize- and enumerate-environments.
    """

    pattern = r'<(ul|ol|dl)(.*)</\1>'
    listing_tag_mapping = {
        'ul': 'itemize',
        'ol': 'enumerate',
        'dl': 'description',
    }

    def __init__(self, *args, **kwargs):
        super(ListConverter, self).__init__(*args, **kwargs)
        self.nesting_level = 0

    def __call__(self):
        html = self.get_html()

        # minidom hates htmlentities, but loves xmlentities -.-
        html = '<dummy>%s</dummy>' % html
        html = html2xmlentities(html)

        # parse DOM
        try:
            dom = minidom.parseString(html)
        except ExpatError, exc:
            # cleanup html with BeautifulSoup
            html = str(BeautifulSoup(html))
            dom = minidom.parseString(html)

        latex = []

        for node in dom.getElementsByTagName('dummy')[0].childNodes:
            if node.nodeType == minidom.Node.ELEMENT_NODE and \
                    node.tagName.lower() in self.listing_tag_mapping.keys():

                latex.extend(self.convert_listing_environment(node))

            else:
                latex.append(self.converter.convert(node.toxml()))

        latex.append('')
        self.replace_and_lock('\n'.join(latex))

    def convert_listing_environment(self, node):
        """Converts a <ul>, <ol> or <dl> node to latex.
        """
        has_items = self._listing_has_items(node)

        if has_items and not self.nesting_level >= LIST_NESTING_LIMIT:
            self.nesting_level += 1
            result = self._convert_reduced_listing_environment(node)
            self.nesting_level -= 1
            return result

        else:
            return self._convert_reduced_listing_environment(
                node, environment=False)

    def _listing_has_items(self, node):
        for elm in node.childNodes:
            if elm.nodeType == minidom.Node.ELEMENT_NODE and \
                    elm.tagName.lower() in ('li', 'dt', 'dd'):
                return True
        return False

    def _convert_reduced_listing_environment(self, node, environment=True):
        """Internal method should only be called
        by ``convert_listing_environment``.
        """
        if node.tagName.lower() in ('ol', 'ul'):
            nodes_latex = self._convert_listing_items(node)

        else:
            nodes_latex = self._convert_description_items(node)

        if not nodes_latex:
            return []

        elif environment:
            begin_env, end_env = self._create_environ(node)
            return ['', begin_env, nodes_latex, end_env]

        else:
            return [nodes_latex]

    def _convert_listing_items(self, list_node):
        """Converts <li> nodes to LaTeX.
        """
        latex = []
        for elm in list_node.childNodes:
            if elm.nodeType == minidom.Node.ELEMENT_NODE and \
                    elm.tagName.lower() == 'li':
                content = self._get_node_content(elm)
                if content:
                    latex.append(r'\item %s' % content.strip())

            elif elm.nodeType == minidom.Node.ELEMENT_NODE and \
                    elm.tagName.lower() in self.listing_tag_mapping.keys():
                latex.extend(self.convert_listing_environment(elm))

            else:
                content_latex = self._get_node_content(elm)
                if content_latex is not None:
                    latex.append(content_latex)

        return '\n'.join(latex)

    def _convert_description_items(self, list_node):
        """Converts <dt> / <dd> nodes to LaTeX.
        """
        latex = []

        dt_node = None
        for elm in list_node.childNodes:
            if elm.nodeType == minidom.Node.ELEMENT_NODE and \
                    elm.tagName.lower() == 'dt':
                dt_node = elm

            elif elm.nodeType == minidom.Node.ELEMENT_NODE and \
                    elm.tagName.lower() == 'dd' and \
                    dt_node is not None:
                dt_content = self._get_node_content(dt_node) or ''
                dd_content = self._get_node_content(elm) or ''
                latex.append(r'\item[%s] %s' % (
                        dt_content.strip(),
                        dd_content.strip()))
                dt_node = None

            elif elm.nodeType == minidom.Node.ELEMENT_NODE and \
                    elm.tagName.lower() in self.listing_tag_mapping.keys():
                latex.extend(self.convert_listing_environment(elm))

            else:
                content_latex = self._get_node_content(elm)
                if content_latex is not None:
                    latex.append(content_latex)

        return '\n'.join(latex)

    def _get_node_content(self, elm):
        """Returns the LaTeX for the node `elm`.
        """
        if elm.nodeType == minidom.Node.TEXT_NODE:
            content_html = elm.toxml().strip()

        else:  # tag node
            content_html = ''.join(
                [e.toxml() for e in elm.childNodes])

        if len(content_html) == 0:
            return None

        else:
            return self.converter.convert(content_html)

    def _create_environ(self, list_):
        """Creates an environment for the node `list_`.
        """
        name = list_.tagName.lower()
        env = self.listing_tag_mapping[name]

        return (r'\begin{%s}' % env,
                r'\end{%s}' % env)





class BuildTerminated(Exception):
    """Raised when trying to run actions (such as add_file) on a closed
    builder.
    """


class PDFBuildFailed(Exception):
    """Raised when pdflatex could not build the PDF for some reason.
    The "message" of the exception should be the stdout of the build
    process.
    """


class ConflictingUsePackageOrder(Exception):
    """The package order is conflicting in the `ILaTeXLayout`.
    """



from ftw.pdfgenerator import interfaces
from zope.interface import implements


class SubConverter(object):
    implements(interfaces.ISubConverter)

    # regular expression
    pattern = None
    placeholder = interfaces.HTML2LATEX_CUSTOM_PATTERN_PLACEHOLDER

    def __init__(self, converter, match, html):
        self.converter = converter
        self.match = match
        self.fullhtml = html

    def __call__(self):
        """This method should be implemented when subclassing.
        """
        raise NotImplementedError()

    def get_html(self):
        return self.fullhtml[self.match.start():self.match.end()]

    def replace_and_lock(self, latex):
        return self.converter.replace_and_lock(
            self.match.start(),
            self.match.end(),
            latex)

    def replace(self, latex):
        return self.converter.replace(
            self.match.start(),
            self.match.end(),
            latex)

    def get_context(self):
        return self.converter.converter.context

    def get_layout(self):
        return self.converter.converter.layout



from ftw.pdfgenerator.customization import LayoutCustomization
from ftw.pdfgenerator.interfaces import ICustomizableLayout
from ftw.pdfgenerator.interfaces import ILayoutCustomization
from ftw.pdfgenerator.layout.customizable import CustomizableLayout
from ftw.pdfgenerator.testing import PDFGENERATOR_ZCML_LAYER
from ftw.pdfgenerator.tests.example import ExampleCustomizableLayout
from ftw.testing import MockTestCase
from zope.interface import Interface
from zope.interface.verify import verifyClass
import os.path


testdata_basedir = os.path.join(os.path.dirname(__file__),
                                'templates')
templates_bar = os.path.join(testdata_basedir, 'bar')
templates_baz = os.path.join(testdata_basedir, 'baz')


class TestCustomizableLayout(MockTestCase):

    layout_class = CustomizableLayout

    layer = PDFGENERATOR_ZCML_LAYER

    def setUp(self, context=None, request=None, builder=None):
        super(TestCustomizableLayout, self).setUp()
        self.context = context or self.create_dummy()
        self.request = request or self.create_dummy()
        self.builder = builder or self.create_dummy()

    def test_implements_interface(self):
        self.replay()
        self.assertTrue(ICustomizableLayout.implementedBy(self.layout_class))
        verifyClass(ICustomizableLayout, self.layout_class)

    def test_template_defines_slots(self):
        self.replay()
        layout = self.layout_class(self.context, self.request, self.builder)

        if not layout.template_name:
            return

        required_slots = [
            'documentclass',
            'usePackages',
            'beneathPackages',
            'aboveDocument',
            'documentTop',
            'documentBottom',
            ]

        template = layout.get_raw_template(layout.template_name)
        self.assertNotEqual(template, None)

        for slot in required_slots:
            self.assertIn('<%%block name="%s"' % slot, template)

    def test_logo_in_template(self):
        self.replay()
        layout = self.layout_class(self.context, self.request, self.builder)

        if not layout.template_name:
            return

        template = layout.get_raw_template(layout.template_name)
        self.assertNotEqual(template, None)
        self.assertIn('${logo}', template)

    def test_rendering_without_customization(self):
        self.replay()
        class Layout(self.layout_class):
            template_directories = [templates_baz]
            template_name = 'example_layout.tex'

        layout = Layout(self.context, self.request, self.builder)
        latex = layout.render_latex('CONTENT')

        self.assertIn(r'\begin{document}', latex)
        self.assertNotIn('my branding', latex)

    def test_rendering_with_customization_and_template(self):
        self.replay()
        class Layout(self.layout_class):
            template_directories = [templates_baz]
            template_name = 'example_layout.tex'

        class Customization(LayoutCustomization):
            template_directories = [templates_bar]
            template_name = 'example_customization.tex'

        self.mock_adapter(Customization, ILayoutCustomization,
                          (Interface, Interface, Interface))

        layout = Layout(self.context, self.request, self.builder)
        latex = layout.render_latex('CONTENT')

        self.assertIn(r'\begin{document}', latex)
        self.assertIn('my branding', latex)

    def test_rendering_with_customization_without_template(self):
        self.replay()
        class Layout(self.layout_class):
            template_directories = [templates_baz]
            template_name = 'example_layout.tex'

        class Customization(LayoutCustomization):
            template_directories = [templates_bar]
            template_name = None

            def before_render_hook(self):
                self.layout.use_package('otherpkg')

            def get_render_arguments(self, args):
                args['logo'] = 'THE-LOGO'
                return args

        self.mock_adapter(Customization, ILayoutCustomization,
                          (Interface, Interface, Interface))

        layout = Layout(self.context, self.request, self.builder)
        latex = layout.render_latex('CONTENT')

        self.assertIn(r'\begin{document}', latex)
        self.assertNotIn('my branding', latex)
        self.assertIn('THE-LOGO', latex)
        self.assertIn(r'\usepackage{otherpkg}', latex)


class TestExampleCustomizableLayout(TestCustomizableLayout):
    layout_class = ExampleCustomizableLayout



from ftw.pdfgenerator.html2latex.subconverters import listing
from ftw.pdfgenerator.tests.base import SubconverterTestBase


class TestListConverter(SubconverterTestBase):

    def test_converter_is_default(self):
        self.assertIn(
            listing.ListConverter,
            self.converter.get_default_subconverters())

    def test_converts_lists_with_win_paths(self):
        self.assertEqual(
            self.convert('<ul><li>O:\\foo\\bar\\baz</li></ul>'),
            '\n\\begin{itemize}\n'
            '\\item O:\\textbackslash foo\\textbackslash bar\\textbackslash baz\n'
            '\\end{itemize}\n')

        self.assertEqual(
            self.convert('<ul><li>One</li><li>Two</li><li>Three</li></ul>'),
            '\n\\begin{itemize}\n\\item One\n\\item Two\n\\item Three\n'
            '\\end{itemize}\n')

        self.assertEqual(
            self.convert('<ol>\n<li>First</li>\n<li>Second</li>\n'
                         '<li>Third</li>\n</ol>'),
            '\n\\begin{enumerate}\n\\item First\n\\item Second\n'
            '\\item Third\n\\end{enumerate}\n')

    def test_ordered_lists(self):
        html = '\n'.join((
                '<ol>',
                '    <li>a</li>',
                '    <li>b',
                '        <ol><li>ba</li></ol>',
                '    </li>',
                '    <li>c</li>',
                '</ol>'))

        latex = '\n'.join((
                '',
                '\\begin{enumerate}',
                '\\item a',
                '\\item b         ',
                '\\begin{enumerate}',
                '\\item ba',
                '\\end{enumerate}',
                '\\item c',
                '\\end{enumerate}',
                ''))

        self.assertEqual(self.convert(html), latex)

    def test_unorderd_lists(self):
        html = '\n'.join(('<ul>',
                          '    <li>a</li>',
                          '    <li>b',
                          '        <ul>',
                          '            <li>ba</li>',
                          '            <li>bb</li>',
                          '        </ul>',
                          '    </li>',
                          '    <li>c</li>',
                          '</ul>'))

        latex = '\n'.join((r'',
                           r'\begin{itemize}',
                           r'\item a',
                           r'\item b         ',
                           r'\begin{itemize}',
                           r'\item ba',
                           r'\item bb',
                           r'\end{itemize}',
                           r'\item c',
                           r'\end{itemize}',
                           r''))

        self.assertEqual(self.convert(html), latex)

    def test_list_values_are_converted(self):
        html = '<ul><li>This is <b>important</b>!</ul>'
        latex = '\n'.join((r'',
                           r'\begin{itemize}',
                           r'\item This is \textbf{important}!',
                           r'\end{itemize}',
                           r''))

        self.assertEqual(self.convert(html), latex)

    def test_list_with_attributes_matches(self):
        html = '<ul style="list-style-type: square;" class="simple">' + \
            '<li>Any text</li></ul>'
        latex = '\n'.join((r'',
                           r'\begin{itemize}',
                           r'\item Any text',
                           r'\end{itemize}',
                           r''))

        self.assertEqual(self.convert(html), latex)

    def test_multiple_lists_at_once(self):
        html = '\n'.join((
                '<ul><li>foo</li><li>bar</li></ul>',
                '<b>foobar</b>',
                '<ul><li>foo</li><li>bar</li></ul>'))

        latex = '\n'.join((
                r'',
                r'\begin{itemize}',
                r'\item foo',
                r'\item bar',
                r'\end{itemize}',
                r'',
                r'\textbf{foobar}',
                r'',
                r'',
                r'\begin{itemize}',
                r'\item foo',
                r'\item bar',
                r'\end{itemize}',
                r''))

        self.assertEqual(self.convert(html), latex)

    def test_wrong_match(self):
        html = '<ulwrong><li>foo</li></ulwrong>'

        latex = 'foo'

        self.assertEqual(self.convert(html), latex)

    def test_bad_html_fallback(self):
        html = '\n'.join((
                '',
                '<ul>',
                '<li>foo</li>',
                'bar',
                '</ul>',
                'baz',
                '<dl>',
                '<dt>foo</dt>',
                '<dd>bar</dd>',
                'baz',
                '</dl>'))

        latex = '\n'.join((
                r'',
                r'\begin{itemize}',
                r'\item foo',
                r'bar',
                r'\end{itemize}',
                r' baz ',
                r'\begin{description}',
                r'\item[foo] bar',
                r'baz',
                r'\end{description}',
                r''))

        self.assertEqual(self.convert(html), latex)

    def test_definition_list(self):
        html = '\n'.join((
                '<dl>',
                '<dt>definition term</dt>',
                '<dd>THE definition</dd>',
                '<dt>another term</dt>',
                '<dd>another definition</dd>',
                '</dl>'))

        latex = '\n'.join((
                r'',
                r'\begin{description}',
                r'\item[definition term] THE definition',
                r'\item[another term] another definition',
                r'\end{description}',
                r''))

        self.assertEqual(self.convert(html), latex)

    def test_nested_definition_list(self):
        html = '\n'.join((
                '<dl>',
                '<dt>definition term</dt>',
                '<dd>THE definition</dd>',
                '<dt>nested</dt>',
                '<dd><dl><dt>subterm</dt><dd>subdef</dd></dl></dd>',
                '<dt>another term</dt>',
                '<dd>another definition</dd>',
                '</dl>'))

        latex = '\n'.join((
                r'',
                r'\begin{description}',
                r'\item[definition term] THE definition',
                r'\item[nested] \begin{description}',
                r'\item[subterm] subdef',
                r'\end{description}',
                r'\item[another term] another definition',
                r'\end{description}',
                r''))

        self.assertEqual(self.convert(html), latex)

    def test_no_list_with_no_elements(self):
        html = '\n'.join((
                'foo',
                '<ul></ul>',
                'bar',
                '<dl>',
                '</dl>',
                'baz',
                '<ol> </ol>'))

        latex = 'foo  bar  baz '

        self.assertEqual(self.convert(html), latex)

    def test_bad_nested_unordered_listings(self):
        html = '\n'.join((
                'foo',
                '<ul>',
                '<ul>',
                '<li>bar</li>',
                '</ul>',
                '</ul>',
                'bar'))

        latex = 'foo \n\\begin{itemize}\n\\item bar\n\\end{itemize}\n bar'

        self.assertEqual(self.convert(html), latex)

    def test_bad_nested_definition_listings(self):
        html = '\n'.join((
                'foo',
                '<dl>',
                '<dl>',
                '<dt>bar</dt>',
                '<dd>baz</dd>',
                '</dl>',
                '</dl>',
                'bar'))

        latex = 'foo \n\\begin{description}\n' + \
            '\\item[bar] baz\n\\end{description}\n bar'

        self.assertEqual(self.convert(html), latex)

    def test_outer_nesting(self):
        html = '\n'.join(('<ul>',
                          '    <li>a</li>',
                          '    <ul>',
                          '        <li>ba</li>',
                          '        <li>bb</li>',
                          '    </ul>',
                          '    <li>b</li>',
                          '    <dl>',
                          '        <dt>foo</dt>',
                          '        <dd>bar</dd>',
                          '        <ol>',
                          '            <li>baz</li>',
                          '        </ol>',
                          '    </dl>',
                          '    <li>c</li>',
                          '</ul>'))

        latex = '\n'.join((r'',
                           r'\begin{itemize}',
                           r'\item a',
                           r'',
                           r'\begin{itemize}',
                           r'\item ba',
                           r'\item bb',
                           r'\end{itemize}',
                           r'\item b',
                           r'',
                           r'\begin{description}',
                           r'\item[foo] bar',
                           r'',
                           r'\begin{enumerate}',
                           r'\item baz',
                           r'\end{enumerate}',
                           r'\end{description}',
                           r'\item c',
                           r'\end{itemize}',
                           r''))

        self.assertEqual(self.convert(html), latex)

    def test_limit_list_nesting_to_4(self):
        """LaTeX allows a maximum list nesting of 4, we need to flatten
        further nested lists into level 4.
        """
        html = '\n'.join((
                r'<ul><li>one</li>',
                r'  <ul><li>two</li>',
                r'    <ul><li>three</li>',
                r'      <ul><li>four</li>',
                r'        <ul><li>five</li>',
                r'          <ul><li>six</li>',
                r'          </ul>',
                r'        </ul>',
                r'      </ul>',
                r'    </ul>',
                r'  </ul>',
                r'</ul>',
                ))

        latex = '\n'.join((
                r'',
                r'\begin{itemize}',
                r'\item one',
                r'',
                r'\begin{itemize}',
                r'\item two',
                r'',
                r'\begin{itemize}',
                r'\item three',
                r'',
                r'\begin{itemize}',
                r'\item four',
                r'\item five',
                r'\item six',
                r'\end{itemize}',
                r'\end{itemize}',
                r'\end{itemize}',
                r'\end{itemize}',
                r'',
                ))
        self.assertEqual(self.convert(html), latex)

    def test_ul_does_not_fail_on_empty_item(self):
        html = '\n'.join((
                '<ul>',
                '<li>foo</li>',
                '<li></li>',
                '<li>bar</li>',
                '</ul>'))

        latex = '\n'.join((
                r'',
                r'\begin{itemize}',
                r'\item foo',
                r'\item bar',
                r'\end{itemize}',
                r'',
                ))

        self.assertEqual(self.convert(html), latex)

    def test_dl_does_not_fail_on_empty_item(self):
        html = '\n'.join((
                '<dl>',
                '<dt>foo</dt>',
                '<dd></dd>',
                '<dt></dt>',
                '<dd>bar</dd>',
                '</dl>'))

        latex = '\n'.join((
                r'',
                r'\begin{description}',
                r'\item[foo] ',
                r'\item[] bar',
                r'\end{description}',
                r'',
                ))

        self.assertEqual(self.convert(html), latex)



from Products.CMFCore.utils import getToolByName


LANGUAGE_CODE_TO_OPTION_MAPPING = {
    'af': 'afrikaans',
    'eu': 'basque',
    'br': 'breton',
    'bg': 'bulgarian',
    'ca': 'catalan',
    'hr': 'croatian',
    'cs': 'czech',
    'da': 'danish',
    'nl': 'dutch',
    'en': 'english',
    'en-us': 'USenglish',
    'en-as': 'american',
    'en-gb': 'british',
    'en-ca': 'canadian',
    'en-au': 'australian',
    'en-nz': 'newzealand',
    'eo': 'esperanto',
    'et': 'estonian',
    'fi': 'finnish',
    'fr': 'french',
    'fr-ca': 'canadian',
    'gl': 'galician',
    'de': 'ngerman',
    'de-at': 'naustrian',
    'de-de': 'ngerman',
    'de-ch': 'ngerman',
    'el': 'greek',
    'he': 'hebrew',
    'hu': 'hungarian',
    'is': 'icelandic',
    'ia': 'interlingua',
    'ga': 'irish',
    'it': 'italian',
    'la': 'latin',
    'no': 'nynorsk',
    'pl': 'polish',
    'pt-br': 'brazilian',
    'pt': 'portuges',
    'ro': 'romanian',
    'ru': 'russian',
    'gd': 'scottish',
    'es': 'spanish',
    'sk': 'slovak',
    'sl': 'slovene',
    'sv': 'swedish',
    'sr': 'serbian',
    'tr': 'turkish',
    'uk': 'ukrainian',
    'cy': 'welsh',
    }


def lookup_babel_option(language_code):
    """Returns the babel option for a particular language code.
    Returns `None` if no option is found.
    """

    code = language_code

    if code in LANGUAGE_CODE_TO_OPTION_MAPPING:
        return LANGUAGE_CODE_TO_OPTION_MAPPING[code]

    if '-' in code:
        code = code.split('-')[0]
        if code in LANGUAGE_CODE_TO_OPTION_MAPPING:
            return LANGUAGE_CODE_TO_OPTION_MAPPING[code]

    return None


def get_preferred_babel_option_for_context(context):
    """Returns the preferred babel option (language) for the current context.
    It tries to get the language of the current context, secondary using the
    current preferred language of the current visitor.

    If no babel option is found, `None` is returned.
    """

    context_language_method = getattr(context, 'getLanguage', None)
    if context_language_method:
        language_code = context_language_method()

    else:
        ltool = getToolByName(context, 'portal_languages')
        language_code = ltool.getPreferredLanguage()

    return lookup_babel_option(language_code)



class CustomPatternAtPlaceholderWrapper(object):
    """
    Used for creating a custom pattern at a specific placeholder.
    It is used instead of a normal mode, wrapping the real mode and defining
    where to insert the placeholder. See placeholders in interfaces module.
    """

    def __init__(self, mode, placeholder):
        self.mode = mode
        self.placeholder = placeholder



# pylint: disable=W0212, W0201
# W0212: Access to a protected member of a client class
# W0201: Attribute defined outside __init__

from ftw.pdfgenerator.exceptions import ConflictingUsePackageOrder
from ftw.pdfgenerator.interfaces import IBuilder
from ftw.pdfgenerator.interfaces import ILaTeXLayout
from ftw.pdfgenerator.interfaces import ILaTeXView
from ftw.pdfgenerator.layout.baselayout import BaseLayout
from ftw.pdfgenerator.testing import PDFGENERATOR_ZCML_LAYER
from ftw.testing import MockTestCase
from zope.annotation import IAnnotations
from zope.component import adaptedBy, provideAdapter
from zope.interface import Interface, implements
from zope.interface import alsoProvides, directlyProvides
from zope.interface.verify import verifyClass


class TestBaseLayout(MockTestCase):

    layer = PDFGENERATOR_ZCML_LAYER

    def setUp(self):
        super(TestBaseLayout, self).setUp()
        self.builder = self.create_dummy()
        directlyProvides(self.builder, IBuilder)

    def test_implements_interface(self):
        self.assertTrue(ILaTeXLayout.implementedBy(BaseLayout))
        verifyClass(ILaTeXLayout, BaseLayout)

    def test_adapts(self):
        # BaseLayout should adapt three things (context, request, builder).
        self.assertEquals(len(adaptedBy(BaseLayout)), 3)

    def test_layout_is_annotatable(self):
        layout = BaseLayout(self.create_dummy(), self.create_dummy(),
                            self.builder)
        ann = IAnnotations(layout)
        self.assertNotEqual(ann, None)

    def test_use_package(self):
        layout = BaseLayout(self.create_dummy(), self.create_dummy(),
                            self.builder)

        self.assertEqual(layout.get_packages_latex(), '')

        layout.use_package('inputenc', 'utf8')
        self.assertEqual(layout.get_packages_latex(),
                         '\\usepackage[utf8]{inputenc}\n')

        layout.use_package(u'titlesec', 'noindentafter',
                           insert_after='inputenc')
        self.assertEqual(layout.get_packages_latex(),
                         '\\usepackage[utf8]{inputenc}\n'
                         '\\usepackage[noindentafter]{titlesec}\n')

        layout.use_package('titlesec', 'largestsep')
        self.assertEqual(
            layout.get_packages_latex(),
            '\\usepackage[utf8]{inputenc}\n'
            '\\usepackage[noindentafter, largestsep]{titlesec}\n')

        layout.use_package('titlesec', 'largestsep')
        self.assertEqual(
            layout.get_packages_latex(),
            '\\usepackage[utf8]{inputenc}\n'
            '\\usepackage[noindentafter, largestsep]{titlesec}\n')

        layout.use_package('titlesec', options='', append_options=False)
        self.assertEqual(
            layout.get_packages_latex(),
            '\\usepackage[utf8]{inputenc}\n'
            '\\usepackage{titlesec}\n')

        # clear use-package list
        layout._packages = []

        layout.use_package('hyperref')
        layout.use_package('hyperref', options='colorlinks=false')
        layout.use_package('hyperref')
        self.assertEqual(
            layout.get_packages_latex(),
            '\\usepackage[colorlinks=false]{hyperref}\n')

    def test_use_package_bad_package_name(self):
        layout = BaseLayout(self.create_dummy(), self.create_dummy(),
                            self.builder)

        with self.assertRaises(ValueError) as cm:
            layout.use_package(None)

        self.assertEqual(
            str(cm.exception),
            'Package name should be a string, got None (NoneType)')

        with self.assertRaises(ValueError) as cm:
            layout.use_package(5)

        self.assertEqual(
            str(cm.exception),
            'Package name should be a string, got 5 (int)')

    def test_remove_package(self):
        layout = BaseLayout(self.create_dummy(), self.create_dummy(),
                            self.builder)
        self.assertEqual(layout.get_packages_latex(), '')

        layout.use_package('color')
        layout.use_package('colortbl')
        layout.use_package('dcolumn')

        self.assertEqual(
            layout.get_packages_latex(),
            '\\usepackage{color}\n'
            '\\usepackage{colortbl}\n'
            '\\usepackage{dcolumn}\n')

        layout.remove_package('color')
        self.assertEqual(
            layout.get_packages_latex(),
            '\\usepackage{colortbl}\n'
            '\\usepackage{dcolumn}\n')

        layout.remove_package('*')
        self.assertEqual(layout.get_packages_latex(), '')

    def test_use_package_order(self):
        # use_package in combination with insert_after
        # should also work when the "other" package is not yet registered.

        layout = BaseLayout(self.create_dummy(), self.create_dummy(),
                            self.builder)
        self.assertEqual(layout.get_packages_latex(), '')

        # test insert_after
        layout.use_package('dcolumn', insert_after='color')
        layout.use_package('color')
        layout.use_package('colortbl')

        self.assertEqual(
            layout.get_packages_latex(),
            '\\usepackage{color}\n'
            '\\usepackage{dcolumn}\n'
            '\\usepackage{colortbl}\n')

        layout.remove_package('*')
        self.assertEqual(layout.get_packages_latex(), '')

    def test_use_package_conflicting_order(self):
        # Using use_package with conflicting orders should raise exceptions

        layout = BaseLayout(self.create_dummy(), self.create_dummy(),
                            self.builder)
        self.assertEqual(layout.get_packages_latex(), '')

        # insert_after itself
        with self.assertRaises(ConflictingUsePackageOrder) as cm:
            layout.use_package('foo', insert_after='foo')

        self.assertEqual(
            str(cm.exception),
            'Cannot insert "foo" after itself (insert_after="foo").')

        self.assertEqual(layout.get_packages_latex(), '')

        # recursive conflicting order
        layout.use_package('baz', insert_after='foo')
        layout.use_package('foo', insert_after='bar')
        with self.assertRaises(ConflictingUsePackageOrder) as cm:
            layout.use_package('bar', insert_after='baz')

        self.assertEqual(
            str(cm.exception),
            'Conflicting order: bar after baz after foo after bar.')

    def test_use_babel_with_linguaplone(self):
        context = self.mocker.mock()
        self.expect(context.getLanguage).result(lambda: 'pt-br')

        self.replay()

        layout = BaseLayout(context, self.create_dummy(),
                            self.builder)
        self.assertEqual(layout.get_packages_latex(), '')

        layout.use_babel()
        self.assertEqual(layout.get_packages_latex(),
                         '\\usepackage[brazilian]{babel}\n')

    def test_use_babel_with_preferred_language(self):
        ltool = self.mocker.mock()
        self.mock_tool(ltool, 'portal_languages')
        self.expect(ltool.getPreferredLanguage()).result('en-gb')

        self.replay()

        layout = BaseLayout(self.create_dummy(), self.create_dummy(),
                            self.builder)
        self.assertEqual(layout.get_packages_latex(), '')

        layout.use_babel()
        self.assertEqual(layout.get_packages_latex(),
                         '\\usepackage[british]{babel}\n')

    def test_use_babel_with_particular_option(self):
        layout = BaseLayout(self.create_dummy(), self.create_dummy(),
                            self.builder)
        self.assertEqual(layout.get_packages_latex(), '')

        layout.use_babel('british')
        self.assertEqual(layout.get_packages_latex(),
                         '\\usepackage[british]{babel}\n')

    def test_use_babel_with_multiple_languages(self):
        layout = BaseLayout(self.create_dummy(), self.create_dummy(),
                            self.builder)
        self.assertEqual(layout.get_packages_latex(), '')

        layout.use_babel(['british', 'french', 'latin'])
        self.assertEqual(layout.get_packages_latex(),
                         '\\usepackage[latin, french, british]{babel}\n')

    def test_use_babel_with_unsupported_language(self):
        ltool = self.mocker.mock()
        self.mock_tool(ltool, 'portal_languages')
        self.expect(ltool.getPreferredLanguage()).result('unknown')

        self.replay()

        layout = BaseLayout(self.create_dummy(), self.create_dummy(),
                            self.builder)
        self.assertEqual(layout.get_packages_latex(), '')

        layout.use_babel()
        self.assertEqual(layout.get_packages_latex(),
                         '\\usepackage{babel}\n')

    def test_use_babel_passes_keyword_args(self):
        layout = BaseLayout(self.create_dummy(), self.create_dummy(),
                            self.builder)
        self.assertEqual(layout.get_packages_latex(), '')

        layout.use_package('babel', options='foo')
        self.assertEqual(layout.get_packages_latex(),
                         '\\usepackage[foo]{babel}\n')

        layout.use_babel('bar')
        self.assertEqual(layout.get_packages_latex(),
                         '\\usepackage[bar]{babel}\n')

        layout.use_babel('baz', append_options=True)
        self.assertEqual(layout.get_packages_latex(),
                         '\\usepackage[bar, baz]{babel}\n')

    def test_render_latex_raises_not_implemented(self):
        # The BaseLayout does not specify how to render the LaTeX.
        layout = BaseLayout(self.create_dummy(), self.create_dummy(),
                            self.builder)

        with self.assertRaises(NotImplementedError):
            layout.render_latex('latex')

    def test_get_builder_returns_builder(self):
        context = self.create_dummy()
        request = self.create_dummy()
        builder = self.create_dummy()
        layout = BaseLayout(context, request, builder)

        self.assertEqual(layout.get_builder(), builder)

    def test_get_converter_keeps_instance(self):
        context = self.create_dummy()
        request = self.create_dummy()
        builder = self.create_dummy()

        layout = BaseLayout(context, request, builder)

        self.assertNotEqual(layout.get_converter(), None)
        self.assertEqual(layout.get_converter(), layout.get_converter())

    def test_get_view_rendering(self):
        class IFoo(Interface):
            pass

        foo = self.create_dummy()
        alsoProvides(foo, IFoo)

        class FooView(object):
            implements(ILaTeXView)
            data = 'foo'

            def __init__(self, context, request, layout):
                pass

            def render(self):
                return self.data
        provideAdapter(FooView, (IFoo, Interface, Interface), name='')

        class PreFooView(FooView):
            data = 'pre'
        provideAdapter(PreFooView, (IFoo, Interface, Interface),
                       name='pre-hook')

        class PostFooView(FooView):
            data = 'post'
        provideAdapter(PostFooView, (IFoo, Interface, Interface),
                       name='post-hook')

        context = self.create_dummy()
        request = self.create_dummy()
        builder = self.create_dummy()
        layout = BaseLayout(context, request, builder)

        views = layout.get_views_for(foo)
        self.assertEqual(len(views), 3)
        self.assertEqual(type(views[0]), PreFooView)
        self.assertEqual(type(views[1]), FooView)
        self.assertEqual(type(views[2]), PostFooView)

        self.assertEqual(
            layout.render_latex_for(foo), 'pre\nfoo\npost')



from ftw.pdfgenerator.interfaces import ICustomizableLayout
from ftw.pdfgenerator.interfaces import ILayoutCustomization
from ftw.pdfgenerator.templating import MakoTemplating
from zope.component import adapts
from zope.interface import Interface
from zope.interface import implements


class LayoutCustomization(MakoTemplating):
    """For documententation see ``ILayoutCustomization``.
    When subclassing, inherit and run the testcase
    ``ftw.pdfgenerator.tests.test_customization.TestLayoutCustomization``.
    """

    implements(ILayoutCustomization)
    adapts(Interface, Interface, ICustomizableLayout)

    template_directories = []
    template_name = None

    def __init__(self, context, request, layout):
        super(LayoutCustomization, self).__init__()
        self.context = context
        self.request = request
        self.layout = layout

    def before_render_hook(self):
        pass

    def get_render_arguments(self, args):
        return args

    def get_template_directories(self):
        dirs = super(LayoutCustomization, self).get_template_directories()
        dirs += self.layout.get_template_directories()
        return dirs

    @property
    def template_lookup(self):
        if getattr(self, '_mako_template_lookup', None) is None:
            lookup = super(LayoutCustomization, self).template_lookup
            lookup.put_template(
                'original_layout',
                lookup.get_template(self.layout.template_name))

        return super(LayoutCustomization, self).template_lookup

    def add_raw_template_file(self, name):
        """Adds the contents of a file to the builder without parsing it.
        """
        self.layout.builder.add_file(name, data=self.get_raw_template(name))



from ftw.pdfgenerator import interfaces
from ftw.pdfgenerator.browser.views import ExportPDFView
from ftw.pdfgenerator.interfaces import DEBUG_MODE_SESSION_KEY
from ftw.pdfgenerator.testing import PDFGENERATOR_ZCML_LAYER
from ftw.testing import MockTestCase
from mocker import ANY
from zope.component import getMultiAdapter
from zope.interface import Interface, directlyProvides
from zope.publisher.interfaces.browser import IDefaultBrowserLayer


class TestAsPDFView(MockTestCase):

    layer = PDFGENERATOR_ZCML_LAYER

    def mock_allow_alternate_output(self, result_value, debug_mode=False):
        request = self.mocker.mock()
        context = self.mocker.mock()
        user = self.mocker.mock()
        self.expect(
            context.portal_membership.getAuthenticatedMember()).result(user)

        self.expect(user.has_permission('cmf.ManagePortal', context)).result(
            result_value)

        session = self.mocker.mock(count=False)
        self.expect(request.SESSION).result(session).count(0, None)
        self.expect(session.get(DEBUG_MODE_SESSION_KEY, False)).result(
            debug_mode)

        return context, request

    def test_component_registered(self):
        request = self.create_dummy()
        directlyProvides(request, IDefaultBrowserLayer)
        aspdf = getMultiAdapter((object(), request), name='export_pdf')
        self.assertTrue(isinstance(aspdf, ExportPDFView))

    def test_get_build_arguments(self):
        context = object()
        request = object()
        aspdf = ExportPDFView(context, request)
        self.assertEqual(aspdf.get_build_arguments(), {'request': request})

    def test_export_as_pdf(self):
        context = object()
        request = object()

        assembler = self.mocker.mock()
        self.mock_adapter(assembler, interfaces.IPDFAssembler,
                          (Interface, Interface))
        self.expect(assembler(ANY, ANY)).result(assembler)
        self.expect(assembler.build_pdf(
                request=request)).result(request)

        self.replay()

        aspdf = ExportPDFView(context, request)
        self.assertEqual(aspdf.export(output='pdf'), request)

    def test_export_as_latex(self):
        context = object()
        request = object()

        assembler = self.mocker.mock()
        self.mock_adapter(assembler, interfaces.IPDFAssembler,
                          (Interface, Interface))
        self.expect(assembler(ANY, ANY)).result(assembler)
        self.expect(assembler.build_latex(
                request=request)).result(request)

        self.replay()

        aspdf = ExportPDFView(context, request)
        self.assertEqual(aspdf.export(output='latex'), request)

    def test_export_as_zip(self):
        context = object()
        request = object()

        assembler = self.mocker.mock()
        self.mock_adapter(assembler, interfaces.IPDFAssembler,
                          (Interface, Interface))
        self.expect(assembler(ANY, ANY)).result(assembler)
        self.expect(assembler.build_zip(
                request=request)).result(request)

        self.replay()

        aspdf = ExportPDFView(context, request)
        self.assertEqual(aspdf.export(output='zip'), request)

    def test_export_with_unkown_output(self):
        context = object()
        request = object()

        assembler = self.mocker.mock()
        self.mock_adapter(assembler, interfaces.IPDFAssembler,
                          (Interface, Interface))
        self.expect(assembler(ANY, ANY)).result(assembler)

        self.replay()

        aspdf = ExportPDFView(context, request)

        with self.assertRaises(ValueError) as cm:
            self.assertEqual(aspdf.export(output='foo'), request)

        self.assertEqual(str(cm.exception),
                         'Unkown output "foo"')

    def test_allow_alternate_output_True(self):
        context, request = self.mock_allow_alternate_output(True)

        self.replay()

        aspdf = ExportPDFView(context, request)
        self.assertEqual(aspdf.allow_alternate_output(), True)

    def test_allow_alternate_output_False(self):
        context, request = self.mock_allow_alternate_output(False)

        self.replay()

        aspdf = ExportPDFView(context, request)
        self.assertEqual(aspdf.allow_alternate_output(), False)

    def test_allow_alternate_output_in_debug_mode_False(self):
        context, request = self.mock_allow_alternate_output(False, True)

        self.replay()

        aspdf = ExportPDFView(context, request)
        self.assertEqual(aspdf.allow_alternate_output(), True)

    def test_call_renders_template_if_admin(self):
        context, request = self.mock_allow_alternate_output(True)

        aspdf = self.mocker.patch(ExportPDFView(context, request),
                                  spec=False)
        self.expect(request.get('submitted', False)).result(False)
        self.expect(aspdf.index()).result('rendered html')

        self.replay()

        self.assertEqual(aspdf(), 'rendered html')

    def test_call_exports_if_not_admin(self):
        context, request = self.mock_allow_alternate_output(False)

        aspdf = self.mocker.patch(ExportPDFView(context, request))
        self.expect(aspdf.export()).result('pdf')

        self.replay()

        self.assertEqual(aspdf(), 'pdf')

    def test_call_uses_output_from_request_if_admin(self):
        context, request = self.mock_allow_alternate_output(True)
        self.expect(request.get('submitted', False)).result(True)
        self.expect(request.get('output')).result('latex')

        aspdf = self.mocker.patch(ExportPDFView(context, request))
        self.expect(aspdf.export('latex')).result('latex code')

        self.replay()

        self.assertEqual(aspdf(), 'latex code')



from ftw.pdfgenerator.browser.views import ExportPDFView
from ftw.pdfgenerator.interfaces import IBuilderFactory
from ftw.pdfgenerator.layout.makolayout import MakoLayoutBase
from ftw.pdfgenerator.view import MakoLaTeXView
from zope.component import getUtility


class BaseStandalonePDFView(ExportPDFView, MakoLayoutBase, MakoLaTeXView):
    """A special standalone PDF view, combining:

    - The export_pdf wizard (ExportPDFView)
    - A LaTeX layout (MakoLayoutBase)
    - A LaTeX view (MakoLaTeXView)

    Extend this class and register it as browser for having a custom PDF
    export which is not recursive. With this approach it is possible to have
    multiple, completely different PDF exports for one single context.
    """

    def __init__(self, context, request):
        # call init on BrowserView, which is the superclass of ExportPDFView
        # pylint: disable=E1003
        # E1003: Bad first argument 'ExportPDFView' given to super class
        super(ExportPDFView, self).__init__(context, request)

        # initialize layout, but with an emty builder. It will be set
        # in get_build_arguments.
        builder = None
        MakoLayoutBase.__init__(self, context, request, builder)

        # initialize the latex view
        layout = self
        MakoLaTeXView.__init__(self, context, request, layout)

    def get_build_arguments(self):
        # ExportPDFView.get_build_arguments
        args = ExportPDFView.get_build_arguments(self)

        # we need to manually create the builder, so that we have a reference
        # to it.
        self.builder = getUtility(IBuilderFactory)()

        # this object is also the layout - tell the assembler
        args.update({'layout': self,
                     'builder': self.builder})

        return args

    def get_views_for(self, obj):
        # MakoLayoutBase.get_views_for

        # Let the layout-part of this object known that this object is
        # also the latex view.
        # We are disable the pre- and post-hooks here!
        return [self]

    def render(self):
        # MakoLaTeXView.render

        # We do not render the view, we use the layout-template for rendering
        # the layout AND the content.
        return ''



from BeautifulSoup import BeautifulSoup
from ftw.pdfgenerator import interfaces
from ftw.pdfgenerator.html2latex import subconverter
from ftw.pdfgenerator.html2latex import wrapper
from ftw.pdfgenerator.html2latex.utils import generate_manual_caption
from ftw.pdfgenerator.utils import html2xmlentities
from operator import methodcaller
from xml.dom import minidom
import re


_marker = object()


MODE_REPLACE = interfaces.HTML2LATEX_MODE_REPLACE
MODE_REGEXP = interfaces.HTML2LATEX_MODE_REGEXP
PLACEHOLDER_BOTTOM = interfaces.HTML2LATEX_CUSTOM_PATTERN_PLACEHOLDER_BOTTOM
PREVENT_CHARACTER = interfaces.HTML2LATEX_PREVENT_CHARACTER
LONGTABLE_ROWS_THRESHOLD = 15

# Global border options:

BORDER_TABLE_L = 2 ** 0
BORDER_TABLE_R = 2 ** 1
BORDER_TABLE_T = 2 ** 2
BORDER_TABLE_B = 2 ** 3
BORDER_TABLE = BORDER_TABLE_L | BORDER_TABLE_R | \
    BORDER_TABLE_T | BORDER_TABLE_B

BORDER_CELL_L = 2 ** 10
BORDER_CELL_R = 2 ** 11
BORDER_CELL_T = 2 ** 12
BORDER_CELL_B = 2 ** 13
BORDER_CELL = BORDER_CELL_L | BORDER_CELL_R | BORDER_CELL_T | BORDER_CELL_B

NO_LAYOUT = 0
GRID_LAYOUT = BORDER_TABLE | BORDER_CELL
HORIZONTAL_LAYOUT = BORDER_TABLE_T | BORDER_TABLE_B | \
    BORDER_CELL_T | BORDER_CELL_B


def parse_html_style_attribute(value):
    data = {}
    if not value:
        return data

    for arg in value.split(';'):
        arg = arg.strip()
        if not arg:
            continue

        if ':' not in arg:
            continue

        key, value = arg.split(':', 1)
        data[key.strip()] = value.strip()

    return data


class TableConverter(subconverter.SubConverter):
    """The TableConverter converts <table>-Tags to latex.
    """

    pattern = r'<table(.*?)>(.*?)</table>'

    def __init__(self, *args, **kwargs):
        super(TableConverter, self).__init__(*args, **kwargs)
        self.columnAmount = 0
        self.LatexColumn = LatexColumn
        self.dom = None
        self.rows = []
        self.columns = []
        self._css_classes = None
        self._environment = None

    def __call__(self):
        self.parse()

        # render table
        latex = self.render()

        # replace with latex
        self.replace_and_lock(latex)

        # register packages
        if self.environment == 'longtable':
            self.converter.converter.layout.use_package('longtable')
        self.converter.converter.layout.use_package('multirow')
        self.converter.converter.layout.use_package('multicol')
        # The "calc" package allows to use "length+length" inline.
        self.converter.converter.layout.use_package('calc')

    def parse(self):
        html = self.get_html()
        # cleanup html with BeautifulSoup
        html = str(BeautifulSoup(html, fromEncoding='utf-8'))
        # minidom hates htmlentities, but loves xmlentities -.-

        html = html2xmlentities(html)

        # parse DOM
        self.dom = minidom.parseString(html)
        self.parse_dom()

    @property
    def environment(self):
        if getattr(self, '_environment', None) is None:
            if 'page-break' in self.get_css_classes():
                self._environment = 'longtable'
            elif 'no-page-break' in self.get_css_classes():
                self._environment = 'tabular'
            elif len(self.rows) > LONGTABLE_ROWS_THRESHOLD:
                self._environment = 'longtable'
            else:
                self._environment = 'tabular'

        return self._environment

    def get_css_classes(self):
        if self._css_classes is None:
            self._css_classes = []
            domTable = self.dom.getElementsByTagName('table')[0]
            classes = domTable.getAttribute('class').strip()
            self._css_classes = classes.split(' ')
        return self._css_classes

    def get_table_layout(self):
        domTable = self.dom.getElementsByTagName('table')[0]

        border = domTable.hasAttribute('border')
        if border and int(border) > 0:
            return GRID_LAYOUT

        elif 'border-grid' in self.get_css_classes() or \
                'grid' in self.get_css_classes():
            return GRID_LAYOUT

        elif 'listing' in self.get_css_classes():
            return HORIZONTAL_LAYOUT

        else:
            return NO_LAYOUT

    def render(self):
        latex = [
            r'\makeatletter\@ifundefined{tablewidth}{' +
            r'\newlength\tablewidth}\makeatother',
            r'\setlength\tablewidth\linewidth',
            r'\addtolength\tablewidth{-%i\tabcolsep}' % (
                2 * len(self.columns)),
            r'\renewcommand{\arraystretch}{1.4}']

        caption_command, insert_caption_at_top = self.render_caption()
        if caption_command and insert_caption_at_top:
            latex.append(caption_command.strip())
            latex.append(r'\vspace{-\baselineskip}')

        latex.append(r'\begin{%s}{%s}' % (
                self.environment, self.get_table_format()))
        latex.append(self.render_rows().strip())
        latex.append(r'\end{%s}\\' % self.environment)
        latex.append(r'\vspace{4pt}')

        if caption_command and not insert_caption_at_top:
            latex.append(r'\vspace{-\baselineskip}')
            latex.append(caption_command.strip())

        latex.append('')
        return '\n'.join(latex)

    def render_rows(self):
        head_rows = []
        body_rows = []

        for row in self.rows:
            rowLatex = row.render()
            if row.is_head_row():
                head_rows.append(rowLatex)
            else:
                body_rows.append(rowLatex)

        head_latex = ''
        if len(head_rows) > 0:
            head_latex = ''.join(head_rows)
        body_latex = ''

        if len(body_rows) > 0:
            body_latex = ''.join(body_rows)

        if len(head_rows) > 0 and self.environment == 'longtable':
            return '%s\\endhead\n%s' % (head_latex, body_latex)
        elif len(head_rows) > 0:
            return head_latex + body_latex
        else:
            return body_latex

    def render_caption(self):
        """
        If the HTML-Table contains a <caption>-Tag, this method
        returns the LaTeX-Command for the caption and weather the
        command should be inserted at the top or at the bottom of
        the table.
        """

        caption, insert_at_top = self._get_caption_from_tag()

        if caption is None:
            caption, insert_at_top = self._get_caption_from_summary()

        if caption is None:
            return None, False

        show_in_index = 'notListed' not in self.get_css_classes()

        latex = generate_manual_caption(caption, 'table',
                                        show_in_index=show_in_index)

        return latex, insert_at_top

    def _get_caption_from_tag(self):
        caption_tags = self.dom.getElementsByTagName('caption')
        if len(caption_tags) == 0:
            return None, False

        caption_tag = caption_tags[0]

        content = ''.join([child.toxml() for child
                           in caption_tag.childNodes])
        content = content.encode('utf8')
        latexContent = self.converter.convert(content)

        siblings = caption_tag.parentNode.childNodes
        insert_at_top = len(siblings) / float(2) > (
            siblings.index(caption_tag) + 1)

        return latexContent, insert_at_top

    def _get_caption_from_summary(self):
        domTable = self.dom.getElementsByTagName('table')[0]

        caption = domTable.getAttribute('summary') or None
        if caption is not None:
            caption = self.converter.convert(caption)

        insert_at_top = False
        return caption, insert_at_top

    def get_table_format(self):
        column_format = []
        for col in self.columns:
            column_format.append(col.get_format())
        return ''.join(column_format)

    def parse_dom(self):
        # create colum objects
        dom_col_list = self.dom.getElementsByTagName('col')
        amount = self.get_amount_of_columns()
        amount = amount and amount or len(dom_col_list)

        for c in range(amount):
            col = self.create_latex_column()
            if len(dom_col_list) > c:
                col.set_dom_col(dom_col_list[c])
            self.columns.append(col)

        # create rows / cells
        multi_row_cache = {
            # columnIndex : [Cell, DecrementingRowspan],
            }

        for domTr in self.dom.getElementsByTagName('tr'):
            self._parse_tr_dom(domTr, multi_row_cache, amount)

    def _parse_tr_dom(self, domTr, multi_row_cache, amount_of_columns):
        row = self.create_latex_row(domTr=domTr)
        self.rows.append(row)

        # .. cells
        cells = domTr.getElementsByTagName('th')
        cells += domTr.getElementsByTagName('td')

        # sorting
        cells.sort(lambda a, b: cmp(a.parentNode.childNodes.index(a),
                                    b.parentNode.childNodes.index(b)))
        columnIndex = 0
        cell_index = 0

        while sum(map(methodcaller('get_colspan'), row.cells)) < \
                amount_of_columns:
            if columnIndex in multi_row_cache.keys():
                multi_row_cache[columnIndex][0].register_row(row)
                multi_row_cache[columnIndex][1] -= 1
                if multi_row_cache[columnIndex][1] == 0:
                    del multi_row_cache[columnIndex]
                columnIndex += 1

            else:
                try:
                    dom_cell = cells[cell_index]
                except IndexError:
                    break
                cell_index += 1
                cell = self.create_latex_cell(dom_cell=dom_cell, row=row)
                # rowspan?
                if cell.get_rowspan() > 1:
                    multi_row_cache[columnIndex] = [
                        cell,
                        cell.get_rowspan() - 1]

                # colspan / find and register columns
                # find and register columns
                try:
                    colspan = int(dom_cell.getAttribute('colspan'))
                except ValueError:
                    colspan = 1

                for cid in range(columnIndex, columnIndex + colspan):
                    while len(self.columns) <= cid:
                        col = self.create_latex_column()
                        self.columns.append(col)
                    cell.register_column(self.columns[cid])
                # move columnIndex to right
                columnIndex += colspan

        return row, columnIndex

    def get_amount_of_columns(self):
        columns = 0
        # by <td>- and <th>-Tags
        for tr in self.dom.getElementsByTagName('tr'):
            tr_columns = 0
            cells = tr.getElementsByTagName('td') + \
                tr.getElementsByTagName('th')
            for cell in cells:
                try:
                    tr_columns += int(cell.getAttribute('colspan'))
                except (ValueError, TypeError):
                    tr_columns += 1
            if tr_columns > columns:
                columns = tr_columns
        return columns

    def create_latex_column(self, *args, **kwargs):
        return LatexColumn(self.converter, self, *args, **kwargs)

    def create_latex_row(self, *args, **kwargs):
        return LatexRow(self.converter, self, *args, **kwargs)

    def create_latex_cell(self, *args, **kwargs):
        return LatexCell(self.converter, self, *args, **kwargs)


class LatexColumn(object):
    """
    Represents a table column (col)
    """

    converter = None
    table_converter = None
    dom_col = None
    cells = []

    def __init__(self, converter, table_converter):
        self.converter = converter
        self.dom_col = None
        self.table_converter = table_converter
        self.cells = []
        self._width = _marker
        self._align = _marker

    def set_dom_col(self, dom_col):
        self.dom_col = dom_col

    def register_cell(self, cell):
        self.cells.append(cell)

    def get_width(self):
        if self._width == _marker:
            self._width = None
            # use <col> definition, if existing
            if self.dom_col:
                try:
                    self._width = LatexWidth.convert(
                        self.dom_col.getAttribute('width'))
                except ValueError:
                    pass

            # .. otherwise try to get width of a cell
            if not self._width:
                for cell in self.cells:
                    w = cell.get_width()
                    if cell.get_colspan() == 1 and w:
                        self._width = w
                        break
        return self._width

    def get_align(self):
        if self._align == _marker:
            if self.dom_col:
                self._align = get_alignment(self.dom_col)
            else:
                self._align = None
        return self._align

    def get_format(self):
        format_ = ''
        if self.get_width():
            format_ = 'p{%s}' % str(self.get_width())

        elif self.get_align():
            mapping = {
                'left': 'l',
                'right': 'r',
                'center': 'c',
                }
            format_ = mapping[self.get_align()]

        else:
            format_ = 'l'

        format_ = apply_borders_to_format(self, format_)
        return format_

    def has_left_border(self):
        return self.table_converter.get_table_layout() & BORDER_CELL_L

    def has_right_border(self):
        if self.is_last_column():
            return self.table_converter.get_table_layout() & BORDER_CELL_R
        else:
            return False

    def is_last_column(self):
        columns = self.table_converter.columns
        return columns[-1] == self

    def is_first_column(self):
        columns = self.table_converter.columns
        return columns[0] == self


class LatexRow(object):
    """
    Represents a table row (tr)
    """

    converter = None
    table_converter = None
    domTr = None
    cells = []

    def __init__(self, converter, table_converter, domTr):
        self.converter = converter
        self.table_converter = table_converter
        self.domTr = domTr
        self.cells = []

    def register_cell(self, cell):
        self.cells.append(cell)

    def is_head_row(self):
        for cell in self.cells:
            if not cell.is_head_cell():
                return False
        return True

    def is_first_row(self):
        return self.table_converter.rows.index(self) == 0

    def get_next_row(self):
        rows = self.table_converter.rows
        next_row_index = rows.index(self) + 1
        try:
            return rows[next_row_index]
        except IndexError:
            return None

    def render(self):
        latex = []

        if self.is_first_row():
            line_latex = self.get_horizontal_line_latex(self, None)
            if line_latex:
                latex.append(line_latex)

        latex_cells = []
        for cell in self.cells:
            tex = cell.render(self)
            if tex:
                latex_cells.append(tex)
            else:
                latex_cells.append('')

        latex.append('%s \\\\' % ' & '.join(latex_cells))
        line_latex = self.get_horizontal_line_latex(self.get_next_row(), self)
        if line_latex:
            latex.append(line_latex)
        return '\n'.join(latex) + '\n'

    def get_horizontal_line_latex(self, top_row, bottom_row):
        lined_column_indexes = []

        if top_row:
            lined_column_indexes.extend(
                top_row._get_cell_indexes_with_border())

        if bottom_row:
            lined_column_indexes.extend(
                bottom_row._get_cell_indexes_with_border(bottom=True))

        # make unique
        lined_column_indexes = set(lined_column_indexes)

        if len(lined_column_indexes) == len(self.table_converter.columns):
            return r'\hline'

        latex = []
        start = -1
        end = -1
        for index in sorted(lined_column_indexes):
            if start == -1:
                start = end = index
            elif end + 1 < index:
                latex.append(r'\cline{%s-%s}' % (start, end))
                start = end = index
            else:
                end = index

        if start != -1:
            latex.append(r'\cline{%s-%s}' % (start, end))

        return ''.join(latex)

    def _get_cell_indexes_with_border(self, bottom=False):
        """Decides for each cell of this row whether it should have a border
        at the top (or at the bottom if `bottom` is `True`).
        It returns each cells column index (starting with 1) for cells which
        have a border.
        """
        if bottom:
            has_border = lambda cell: cell.has_bottom_border_in_row(self)
        else:
            has_border = lambda cell: cell.has_top_border_in_row(self)

        indexes = []
        for cell in self.cells:
            if not has_border(cell):
                continue

            for column in cell.columns:
                indexes.append(self.table_converter.columns.index(column) + 1)

        return indexes


class LatexCell(object):
    """
    Represents a table cell (td, th)
    """

    converter = None
    table_converter = None
    dom_cell = None
    rows = []
    columns = []

    def __init__(self, converter, table_converter, dom_cell, row):
        self.converter = converter
        self.table_converter = table_converter
        self.dom_cell = dom_cell
        self.rows = []
        self.columns = []
        self._parentNodes = None
        self._width = _marker
        self._colspan = None
        self._rowspan = None
        self._align = _marker
        self._css_classes = None
        self.register_row(row)

    def register_row(self, row):
        row.register_cell(self)
        self.rows.append(row)

    def register_column(self, column):
        self.columns.append(column)
        column.register_cell(self)

    def render(self, row):
        # only render it in first occuring row (if rowspan > 1)
        if self.get_rowspan() > 1 and self.rows.index(row) > 0:
            return None
        # render it
        if self.get_rowspan() > 1:
            span = str(self.get_rowspan())
            latex = '\\multirow{%s}{%s}{%s}' % (
                span,
                str(self.get_calculated_width()),
                self.render_content(),
                )
        else:
            span = str(self.get_colspan())
            latex = '\\multicolumn{%s}{%s}{%s}' % (
                span,
                self.get_format(),
                self.render_content(),
                )
        return latex

    def render_content(self):
        content = ''.join([child.toxml() for child
                           in self.dom_cell.childNodes])

        # Carriage returns are not allowed in table cells with multicolumn.
        # We use \newline instead, which only creates a newline if the cell
        # width is defined, but does not fail otherwise.
        custom_patterns = [
            (MODE_REGEXP, r'<br[ \W]{0,}>\n*',
             r'\%snewline ' % PREVENT_CHARACTER),

            (wrapper.CustomPatternAtPlaceholderWrapper(
                    MODE_REPLACE, PLACEHOLDER_BOTTOM), '\n', r'\newline '),

            ]

        content = content.encode('utf8')
        latex = self.converter.convert(content,
                                       custom_patterns=custom_patterns)

        if 'grey' in self.get_css_classes():
            self.converter.converter.layout.use_package('xcolor')
            latex = r'\textcolor{gray}{%s}' % latex

        if 'footnotesize' in self.get_css_classes():
            latex = r'\footnotesize %s' % latex

        if 'scriptsize' in self.get_css_classes():
            latex = r'\scriptsize %s' % latex

        if 'bold' in self.get_css_classes() or self.is_head_cell():
            latex = r'\textbf{%s}' % latex

        if self.get_calculated_width() and self.get_align():
            # if a width and align is defined, we need to set
            # aligenement in content part with a command
            mapping = {
                'left': r'\raggedright ',
                'right': r'\raggedleft ',
                'center': r'\centering ',
                }

            if self.get_align() in mapping.keys():
                latex = mapping[self.get_align()] + latex

        if 'indent2' in self.get_css_classes():
            latex = r'\hangindent 0.2cm\hspace{0.2cm} %s' % latex
        elif 'indent10' in self.get_css_classes():
            latex = r'\hangindent 1cm\hspace{1cm} %s' % latex

        return latex

    def is_head_cell(self):
        if 'thead' in [p.tagName.lower() for p in self.get_parent_nodes()]:
            # cell is within a <thead>
            return True

        elif self.dom_cell.tagName.lower() == 'th':
            # cell is a <th>
            return True

        elif 'vertical' in self.table_converter.get_css_classes() \
                and self.columns[0].is_first_column():
            # "vertical" class makes first column to be "head" cells.
            return True

        else:
            # not a head cell
            return False

    def get_parent_nodes(self):
        """
        Returns a list of all parent nodes (and their parent nodes etc).
        """
        if self._parentNodes is None:
            self._parentNodes = []
            node = self.dom_cell

            while node.parentNode and \
                    node.parentNode.__class__ != minidom.Document:
                node = node.parentNode
                self._parentNodes.append(node)
        return self._parentNodes

    def get_width(self):
        if self._width == _marker:
            widths = [
                parse_html_style_attribute(
                    self.dom_cell.getAttribute('style')).get('width', ''),
                self.dom_cell.getAttribute('width'),
                ]

            self._width = None

            for width in widths:
                if not width:
                    continue

                try:
                    self._width = LatexWidth.convert(width)
                except ValueError:
                    continue
                else:
                    break

        return self._width

    def get_colspan(self):
        if self._colspan is None:
            try:
                self._colspan = int(self.dom_cell.getAttribute('colspan'))
            except (AttributeError, ValueError):
                self._colspan = 1
        return self._colspan

    def get_rowspan(self):
        if self._rowspan is None:
            try:
                self._rowspan = int(self.dom_cell.getAttribute('rowspan'))
            except (AttributeError, ValueError):
                self._rowspan = 1
        return self._rowspan

    def get_calculated_width(self):
        if self.get_width():
            return self.get_width()

        elif self.get_colspan() == 1:
            return self.columns[0].get_width()

        elif self.get_colspan() > 1:
            width = self.columns[0].get_width()
            for i in range(1, len(self.columns)):
                try:
                    width += self.columns[i].get_width()
                except (TypeError, ValueError):
                    # one of the columns has no or invalid width (not addable)
                    return None
            # Compensate padding of spanned columns
            padding = r'%i\tabcolsep' % (2 * (len(self.columns) - 1))
            width = '+'.join((str(width), padding))
            return width

    def get_align(self):
        if self._align == _marker:
            self._align = None

            if self.dom_cell:
                self._align = get_alignment(self.dom_cell)

            if not self._align and self.get_colspan() == 1:
                self._align = self.columns[0].get_align()
        return self._align

    def has_left_border(self):
        if self.columns[0].has_left_border():
            return True

        if 'border-left' in self.get_css_classes():
            return True

        left_cell = self.get_left_cell()
        if left_cell and left_cell.get_rowspan() > 1 and \
                left_cell.has_right_border():
            return True

        return False

    def has_right_border(self):
        if self.columns[-1].has_right_border():
            return True

        if 'border-right' in self.get_css_classes():
            return True

        right_cell = self.get_right_cell()
        if right_cell and right_cell.get_rowspan() > 1 and \
                right_cell.has_left_border():
            return True

        return False

    def get_left_cell(self, row_index=0):
        """Get the next cell at the left side of this cell or None.
        """
        row_cells = self.rows[row_index].cells
        left_cell_index = row_cells.index(self) - 1
        if left_cell_index < 0:
            return None
        else:
            return row_cells[left_cell_index]

    def has_bottom_border_in_row(self, row):
        if self.rows[-1] != row:
            return False

        if self.table_converter.get_table_layout() & BORDER_CELL_B:
            return True

        if 'border-bottom' in self.get_css_classes():
            return True

        return False

    def has_top_border_in_row(self, row):
        if self.rows[0] != row:
            return False

        if self.table_converter.get_table_layout() & BORDER_CELL_T:
            return True

        if 'border-top' in self.get_css_classes():
            return True

        return False

    def get_right_cell(self, row_index=0):
        """Get the next cell at the right side of this cell or None.
        """
        row_cells = self.rows[row_index].cells
        right_cell_index = row_cells.index(self) + 1
        try:
            return row_cells[right_cell_index]
        except IndexError:
            return None

    def get_format(self):
        format_ = 'l'
        if self.get_calculated_width():
            format_ = 'p{%s}' % str(self.get_calculated_width())

        elif self.get_align():
            mapping = {
                'left': 'l',
                'right': 'r',
                'center': 'c',
                }
            format_ = mapping[self.get_align()]

        # use column definition
        elif self.get_colspan() == 1:
            format_ = self.columns[0].get_format()

        format_ = apply_borders_to_format(self, format_)
        return format_

    def get_css_classes(self):
        if self._css_classes is None:
            classes = self.dom_cell.getAttribute('class').strip()
            self._css_classes = classes.split(' ')

        return self._css_classes


class LatexWidth(object):
    """
    Converts HTML-width to latex width
    """

    TYPE_ABSOLUTE = 'absolute width (px, em, ..)'
    TYPE_RELATIVE = 'relative width (% of tablewidth, ..)'
    VALID_ABSOLUTE_UNITS = [
        'cm',
        'mm',
        'em',
        ]

    def __init__(self, width=0, type_=None, unit=''):
        self.width = width
        self.type = type_
        self.unit = unit

    @classmethod
    def convert(cls, width):
        """
        creates a LatexWidth object by a HTML width-Attribute
        """

        # px is the same as no measure unit, so we can strip it.
        width = width.strip()
        if width.endswith('px'):
            width = width.rstrip('px')

        # absolute
        for unit in cls.VALID_ABSOLUTE_UNITS:
            match = re.compile('^([0-9,\.]{1,})(%s)$' % unit).match(width)
            if match:
                w, u = match.groups()
                w = w.replace(',', '.')
                return cls(width=float(w),
                           type_=LatexWidth.TYPE_ABSOLUTE,
                           unit=u)

        # relative
        unit = '%'
        match = re.compile('^([0-9,\.]{1,})(%s)$' % unit).match(width)
        if match:
            w, u = match.groups()
            w = w.replace(',', '.')
            return cls(width=float(w) / 100,
                       type_=LatexWidth.TYPE_RELATIVE,
                       unit=None)

        # without unit -> use 'em'
        match = re.compile('^([0-9,\.]*)$').match(width)
        if match:
            w = width.replace(',', '.')
            return cls(width=float(w),
                       type_=LatexWidth.TYPE_ABSOLUTE,
                       unit='em')

        raise ValueError(
            'Could not convert string "%s" to valid LatexWidth' % width)

    def get(self):
        if self.type == LatexWidth.TYPE_ABSOLUTE:
            return '%s%s' % (str(self.width), self.unit)
        elif self.type == LatexWidth.TYPE_RELATIVE:
            return '%s\\tablewidth' % str(self.width)

    def __str__(self):
        return self.get()

    def __add__(self, b):
        if not isinstance(b, LatexWidth):
            raise ValueError(
                'Cannot accumulate LatexWidth with %s' % (
                    type(b).__name__))

        a, b = self.__class__.harmonize_units(self, b)

        if a.type != b.type:
            raise ValueError(
                'Cannot accumulate relative and absolute widths.')

        if a.unit != b.unit:
            raise ValueError(
                'Cannot accumulate LatexWidths with different ' +
                'units (%s, %s)' % (a.unit, b.unit))

        return LatexWidth(width=(a.width + b.width),
                          type_=a.type,
                          unit=a.unit)

    @staticmethod
    def harmonize_units(a, b):
        """If possible, converts the units of a and b so that they are
        the same.
        """

        if a.type != b.type or a.unit == b.unit:
            return a, b

        elif a.unit == 'mm' and b.unit == 'cm':
            b2 = LatexWidth(width=b.width * 10,
                            type_=b.type,
                            unit='mm')
            return a, b2

        elif a.unit == 'cm' and b.unit == 'mm':
            a2 = LatexWidth(width=a.width * 10,
                            type_=a.type,
                            unit='mm')
            return a2, b

        else:
            return a, b


def apply_borders_to_format(element, format_):
    """Apply the border information to a column / cell format if necessary
    and return the new format.

    Arguments:
    `element` -- Cell or column object.
    `format_` -- LaTeX format definiton.
    """
    if element.has_left_border() and not format_.startswith('|'):
        format_ = '|%s' % format_

    if element.has_right_border() and not format_.endswith('|'):
        format_ = '%s|' % format_

    return format_


def get_alignment(node):
    """Returns the alignment of a XML node.
    """

    alignment_classes = ('left', 'right', 'center')
    classes = node.getAttribute('class').strip().split(' ')

    for cls in alignment_classes:
        if cls in classes:
            return cls

    return node.getAttribute('align')



# pylint: disable=W0212, W0201
# W0212: Access to a protected member of a client class
# W0201: Attribute defined outside __init__

from ftw.pdfgenerator.html2latex.converter import HTML2LatexConverter
from ftw.testing import MockTestCase


LATEX_FOOTNOTE = r'%(text)s\footnote{%(footnote)s}'


class TestFootnoteConverter(MockTestCase):

    def setUp(self):
        super(TestFootnoteConverter, self).setUp()

        self.context = self.mocker.mock(count=False)
        self.expect(self.context.absolute_url()).result(
            'http://nohost/plone')

        self.layout = self.mocker.mock()

        self.converter = HTML2LatexConverter(
            context=self.context,
            request=object(),
            layout=self.layout)

        self.convert = self.converter.convert

    def test_basic_fn_syntax(self):
        self.replay()
        html = ('<span class="footnote" data-footnote="thefn">'
                'complicated</span>')
        latex = LATEX_FOOTNOTE % {'text': 'complicated',
                                  'footnote': 'thefn'}
        self.assertEqual(self.convert(html), latex)

    def test_fn_with_icon(self):
        self.replay()
        html = ('<span class="footnote" data-footnote="thefn">complicated'
                '<i class="some-icon"></i></span>')
        latex = LATEX_FOOTNOTE % {'text': r'complicated\textit{}',
                                  'footnote': 'thefn'}
        self.assertEqual(self.convert(html), latex)

    def test_fn_with_custom_classes(self):
        self.replay()
        html = ('<span class="classa footnote classb" '
                'data-footnote="thefn">complicated</span>')
        latex = LATEX_FOOTNOTE % {'text': r'complicated',
                                  'footnote': 'thefn'}
        self.assertEqual(self.convert(html), latex)

    def test_fn_with_custom_attributes(self):
        self.replay()
        html = ('<span stuff="ignored" class="footnote" '
                'id="nonrelevant" data-footnote="thefn">complicated</span>')
        latex = LATEX_FOOTNOTE % {'text': r'complicated',
                                  'footnote': 'thefn'}
        self.assertEqual(self.convert(html), latex)

    def test_footnote_without_data_footnote_attribute_is_ignored(self):
        self.replay()
        html = ('<span class="footnote">complicated</span>')
        self.assertEqual(self.convert(html), 'complicated')

    def test_span_without_footnote_is_ignored(self):
        self.replay()
        html = ('<span>complicated</span>')
        self.assertEqual(self.convert(html), 'complicated')

    def test_footnote_with_newlines(self):
        self.replay()
        html = ('<span class="footnote" data-footnote="fn\ntext">'
                'complicated</span>')
        latex = LATEX_FOOTNOTE % {'text': 'complicated',
                                  'footnote': 'fn text'}
        self.assertEqual(self.convert(html), latex)

    def test_footnote_with_umlauts(self):
        self.replay()
        html = (u'<span class="footnote" data-footnote="\xfcbl\xe4">'
                u'c\xf6mpli\xe7ated</span>')
        latex = LATEX_FOOTNOTE % {'text': 'c\xc3\xb6mpli\xc3\xa7ated',
                                  'footnote': '\xc3\xbcbl\xc3\xa4'}
        self.assertEqual(self.convert(html), latex)



from ftw.pdfgenerator.interfaces import IConfig
from zope.interface import implements
import tempfile


class DefaultConfig(object):
    implements(IConfig)

    remove_build_directory = True

    def get_build_directory(self):
        return tempfile.mkdtemp(prefix='ftw.pdfgenerator_')



from ftw.pdfgenerator.interfaces import ILaTeXLayout
from ftw.pdfgenerator.interfaces import ILaTeXView
from ftw.pdfgenerator.interfaces import IRecursiveLaTeXView
from ftw.pdfgenerator.templating import MakoTemplating
from zope.component import adapts
from zope.interface import implements, Interface


class MakoLaTeXView(MakoTemplating):
    """A LaTeXView with Mako templating enabled.
    """

    implements(ILaTeXView)
    adapts(Interface, Interface, ILaTeXLayout)

    template_directories = []
    template_name = None

    def __init__(self, context, request, layout):
        super(MakoLaTeXView, self).__init__()
        self.context = context
        self.request = request
        self.layout = layout

    def register_template_file(self, filename, render=False, **kwargs):
        """Registers a file from the template directories to the builder.
        This copies the file in the directory where the latex will be built.

        Arguments:
        filename -- Name of the file stored in a known template directory.
        render -- Render the template first (default: `False`).
        **kwargs -- Arguments passed to the renderer, if render is `True`.
        """

        if render:
            data = self.render_template(filename, **kwargs)

        else:
            data = self.get_template(filename)

        self.layout.get_builder().add_file(filename, data)

    def get_render_arguments(self):
        """The returned dict of this function will be passed to the template
        when its rendered. This makes it easy to pass additional values.
        """
        return {}

    def render(self):
        """By default, the template `template_name`, which is in one of
        the directories defined in `template_directories`, is rendered
        and returned.
        """

        if self.template_name is None:
            raise ValueError('%s: `template_name` is not defined.' % (
                    self.__class__.__name__))

        return self.render_template(self.template_name,
                                    **self.get_render_arguments())

    def convert(self, *args, **kwargs):
        """Convert HTML to LaTeX using the IHTML2LaTeXConverter.
        See the IHTML2LaTeXConverter.convert documentation.
        """
        return self.layout.get_converter().convert(*args, **kwargs)

    def convert_plain(self, *args, **kwargs):
        """Convert plain text to LaTeX by encoding it with html entities
        and then converting the HTML to latex.
        """
        return self.layout.get_converter().convert_plain(*args, **kwargs)


class RecursiveLaTeXView(MakoLaTeXView):
    """A recursive LaTeX view, which also renders the content of the child
    objects.
    """

    implements(IRecursiveLaTeXView)

    def render_children(self):
        """Render LaTeX views of children and return the LaTeX content.
        """

        latex = []

        # If the export was started on this object and "paths" are passed
        # from folder_contents, we should only export the selected objects.
        if self.context == self.layout.context:
            paths = self.request.get('paths', None)
        else:
            paths = None

        for obj in self.context.listFolderContents():
            if paths and '/'.join(obj.getPhysicalPath()) not in paths:
                continue

            data = self.layout.render_latex_for(obj)
            if data:
                latex.append(data)

        return '\n'.join(latex)

    def get_render_arguments(self):
        args = super(RecursiveLaTeXView, self).get_render_arguments()
        args['latex_content'] = self.render_children()
        return args



from Products.Archetypes.utils import contentDispositionHeader
from ftw.pdfgenerator.interfaces import IBuilderFactory
from ftw.pdfgenerator.interfaces import ILaTeXLayout
from ftw.pdfgenerator.interfaces import IPDFAssembler
from zope.component import adapts
from zope.component import getMultiAdapter, getUtility
from zope.interface import implements, Interface


class PDFAssembler(object):
    implements(IPDFAssembler)
    adapts(Interface, Interface)

    def __init__(self, context, request):
        self.context = context
        self.request = request
        self._layout = None
        self._view = None
        self._builder = None

    def build_pdf(self, layout=None, builder=None, request=None, filename=None):
        self._layout = layout
        self._builder = builder

        latex = self.render_latex()
        data = self.get_builder().build(latex)

        if not request:
            return data

        else:
            return self._attach_to_response(
                request, data, 'pdf', filename=filename)

    def build_latex(self, layout=None, builder=None, request=None, filename=None):
        self._layout = layout
        self._builder = builder

        latex = self.render_latex()
        self.get_builder().cleanup()
        return latex

    def build_zip(self, layout=None, builder=None, request=None, filename=None):
        self._layout = layout
        self._builder = builder

        latex = self.render_latex()
        data = self.get_builder().build_zip(latex).read()

        if not request:
            return data

        else:
            return self._attach_to_response(request, data, 'zip', filename=filename)

    def get_builder(self):
        """Returns the IBuilder instance.
        """
        if getattr(self, '_builder', None) is None:
            self._builder = getUtility(IBuilderFactory)()
        return self._builder

    def get_layout(self):
        """Returns the ILaTeXLayout instance.
        """
        if getattr(self, '_layout', None) is None:
            self._layout = getMultiAdapter(
                (self.context, self.request, self.get_builder()),
                ILaTeXLayout)
        return self._layout

    def render_latex(self):
        """Renders the LaTeX for the configured view and layout.
        """
        layout = self.get_layout()
        content_latex = layout.render_latex_for(self.context)
        return layout.render_latex(content_latex)

    def _attach_to_response(self, request, data, extension, filename=None):
        if not filename:
            filename = self.context.id

        if isinstance(filename, unicode):
            filename = filename.encode('utf-8')

        filename = '{}.{}'.format(filename, extension)

        response = request.RESPONSE
        response.setHeader('Content-Type',
                           'application/%s; charset=utf-8' % extension)

        response.setHeader(
            'Content-disposition',
            contentDispositionHeader(
                'attachment', 'utf-8', filename=filename))

        response.write(data)
        return request



from setuptools import setup, find_packages
import os

version = '1.6.3'

tests_require = [
    'unittest2',
    'mocker',
    'plone.mocktestcase',
    'ftw.testing',
    'plone.testing',

    'zope.annotation',
    'zope.component',
    'zope.i18n',
    'zope.interface',
    'zope.publisher',
    ]

setup(name='ftw.pdfgenerator',
      version=version,
      description="A library for generating PDF representations of Plone " + \
          "objects with LaTeX.",
      long_description=open("README.rst").read() + "\n" + \
          open(os.path.join("docs", "HISTORY.txt")).read(),

      classifiers=[
        "Environment :: Web Environment",
        'Framework :: Plone',
        'Framework :: Plone :: 4.3',
        'Framework :: Plone :: 5.1',
        "Intended Audience :: Developers",
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
        ],

      keywords='ftw pdf generator',
      author='4teamwork AG',
      author_email='mailto:info@4teamwork.ch',
      url='https://github.com/4teamwork/ftw.pdfgenerator',
      license='GPL2',

      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['ftw'],
      include_package_data=True,
      zip_safe=False,

      install_requires=[
        'setuptools',

        # Zope
        'Zope2',
        'zope.annotation',
        'zope.component',
        'zope.i18n',
        'zope.i18nmessageid',
        'zope.interface',

        # Plone
        'Products.Archetypes',
        'Products.CMFCore',

        'Mako',
        'BeautifulSoup!=4.0b',
        # -*- Extra requirements: -*-
        ],
      tests_require=tests_require,
      extras_require=dict(tests=tests_require),

      entry_points="""
      # -*- Entry points: -*-
      [z3c.autoinclude.plugin]
      target = plone
      """,
      )



from ftw.pdfgenerator.interfaces import ITemplating, ILaTeXLayout
from ftw.pdfgenerator.layout.makolayout import MakoLayoutBase
from ftw.testing import MockTestCase
from zope.interface.verify import verifyClass
import os


class TestMakoLayoutBase(MockTestCase):

    def setUp(self):
        super(TestMakoLayoutBase, self).setUp()
        testdata_basedir = os.path.join(
            os.path.dirname(__file__), 'templates')

        self.templates_foo = os.path.join(testdata_basedir, 'foo')

    def test_implements_interfaces(self):
        self.assertTrue(ITemplating.implementedBy(MakoLayoutBase))
        verifyClass(ITemplating, MakoLayoutBase)

        self.assertTrue(ILaTeXLayout.implementedBy(MakoLayoutBase))
        verifyClass(ILaTeXLayout, MakoLayoutBase)

    def test_render_with_arguments(self):
        class FooLayout(MakoLayoutBase):
            template_directories = [self.templates_foo]
            template_name = 'titlepage.tex'

            def get_render_arguments(self):
                return {'title': 'hello world!!!'}

        foo = FooLayout(object(), object(), object())
        self.assertEqual(
            foo.render_latex(''),
            '\\title{hello world!!!}\n\n\\maketitle\n\n')

    def test_default_render_arguments(self):
        layout = MakoLayoutBase(object(), object(), object())
        self.assertEqual(layout.get_render_arguments(), {})

    def test_fails_if_no_default_template(self):
        class BarLayout(MakoLayoutBase):
            pass

        bar = BarLayout(object(), object(), object())

        with self.assertRaises(ValueError) as cm:
            bar.render_latex('')

        self.assertEqual(
        str(cm.exception),
        'BarLayout: `template_name` is not defined.')

    def test_add_raw_template_file(self):
        class FooLayout(MakoLayoutBase):
            template_directories = [self.templates_foo]

        builder = self.mocker.mock()
        self.expect(builder.add_file(
                'welcome.tex', data='{\\large Hello \\textbf{${view.name}}!}\n'))

        self.replay()

        foo = FooLayout(object(), object(), builder)
        foo.add_raw_template_file('welcome.tex')



from Products.CMFCore.utils import getToolByName
from ftw.pdfgenerator.html2latex import subconverter
import os.path
import re


class HyperlinkConverter(subconverter.SubConverter):

    pattern = r'<a.*?href="(.*?)".*?>(.*?)</a>'

    def __call__(self):
        context = self.get_context()

        url, label = self.match.groups()
        label = (self.converter.convert(label)
                 .replace('""/', '/')
                 .replace('_', '\_')
                 .replace('"=', '-'))

        is_relative = '://' not in url and not url.startswith('mailto:')

        if is_relative:
            url = os.path.join(context.absolute_url(), url)
            url = url.replace('/./', '/')

        url = self.resolve_uid(url)

        url = url.replace('&amp;', '&')
        url = url.replace('&', '\\&')
        url = url.replace(' ', '%20').replace('%', '\%')
        url = url.replace('#', '\#')

        # Do not display "mailto:"
        url_label = re.sub('^mailto:', '', url)
        url_label = url_label.replace('_', '\_')
        url_label = url_label.replace('""/', '/')

        self.get_layout().use_package('url', 'hyphens')
        self.get_layout().use_package('hyperref')
        self.replace_and_lock(self.latex_link(url, label, url_label))

    def latex_link(self, url, label, url_label):
        href = r'\href{%s}{%s}'
        footnote = r'\footnote{%s}' % href % (url, r'\url{%s}' % url_label)
        return href % (url, label + footnote)

    def resolve_uid(self, url):
        if '/' not in url:
            return url

        parts = url.split('/')
        if parts[-2] == 'resolveuid' or parts[-2] == 'resolveUid':
            context = self.converter.converter.context
            reference_catalog = getToolByName(context, 'reference_catalog')

            uid = parts[-1]
            obj = reference_catalog.lookupObject(uid)

            if obj is not None:
                url = obj.absolute_url()

        return url
